
import numpy as np
import torch
import os
import pandas as pd
import json

from metient.util import vertex_labeling_util as vutil
from metient.util import data_extraction_util as dutil

def write_pairtree_inputs(patient_data, patient_id, output_dir):
    '''
    Args:
        patient_data: pandas DataFrame with at least columns: [character_label, character_index
        sample_label, var, ref, major_cn, minor_cn]. character_label = name of mutation (e.g. "KCTD15:19:34584166:C"),
        character_index = unique index for each character label, sample_label = name of the sample,
        (e.g. "tumor1_region1"), var = variant allele read count, ref = reference allele read count
        
        patient_id: used to name the output files
        
    Outputs:
        Saves ssm and params.json files to output_dir as needed to run PairTree and/or Orchard. 
        See "Input Files" section here: https://github.com/morrislab/pairtree
    
    '''
    header = ["id", "name", "var_reads", "total_reads", "var_read_prob"]

    mutation_names = list(patient_data['character_label'].unique())
    mutation_ids  = list(patient_data['character_index'].unique())
    sample_names = list(patient_data['sample_label'].unique())

    mut_name_to_mut_id = {}
    with open(os.path.join(output_dir, f"{patient_id}.ssm"), "w") as f:

        f.write("\t".join(header))
        f.write("\n")
        for i, mut in zip(mutation_ids, mutation_names):
            mut_name_to_mut_id[mut] = f"m{i}"
            row = [f"m{i}", mut]
            mut_patient_subset = patient_data[patient_data['character_label'] == mut]
            var_reads = []
            total_reads = []
            var_read_probs = []
            for sample in sample_names:
                mut_patient_sample = mut_patient_subset[mut_patient_subset['sample_label'] == sample]
                var = int(mut_patient_sample['var'].values[0])
                ref = int(mut_patient_sample['ref'].values[0])
                var_reads.append(str(var))
                total_reads.append(str(var+ref))
                major_cn = int(mut_patient_sample['major_cn'].values[0])
                minor_cn = int(mut_patient_sample['minor_cn'].values[0])
                p = float(mut_patient_sample['purity'].values[0])
                x = (p*(major_cn+minor_cn)+2*(1-p))
                var_read_prob = (p*major_cn)/x
                var_read_probs.append(str(var_read_prob))

            row += [",".join(var_reads), ",".join(total_reads), ",".join(var_read_probs)]
            f.write("\t".join(row))
            f.write("\n")
    json_data = {"samples": sample_names, "clusters": [[mut_name_to_mut_id[name]] for name in mutation_names], "garbage": []}

    with open(os.path.join(output_dir, f"{patient_id}.params.json"), 'w', encoding='utf-8') as f:
        json.dump(json_data, f, ensure_ascii=False)


def _get_cluster_id(mut_id, clusters):
    '''
    Given mut_id = 'm1' and clusters= [['m1'], ['m0', 'm2']], returns 1 
    '''
    for i, cluster in enumerate(clusters):
        if mut_id in cluster:
            return i
    
    print(f"mut id {mut_id} not found")
    return None

def write_clustered_params_json_from_pyclone_clusters(clusters_tsv_fn, pairtree_ssm_fn, pairtree_params_fn, 
                                                      output_dir, patient_id):
    '''
    After clustering with PyClone (see: https://github.com/Roth-Lab/pyclone),
    prepares clustered params.json files needed for PairTree/Orchard tree building

    Args:

        clusters_tsv_fn: PyClone results tsv that maps each mutation to a cluster id

        pairtree_ssm_fn: PairTree .ssm file which maps mutation ids (e.g. m0) to mutation
        names (e.g. SETD2:3:47103798:G)

        output_dir: where to save clustered params.json

        patient_id: name of patient used for saving

    Outputs:

        Saves clustered params.json at {output_dir}/{patient_id}_pyclone_clustered.params.json

    '''

    pyclone_df = pd.read_csv(clusters_tsv_fn, delimiter="\t")
    pt_ssm_df = pd.read_csv(pairtree_ssm_fn, delimiter="\t")
    cluster_id_to_mut_ids = dict() # 3 -> [m0, m5 ...]

    with open(pairtree_params_fn) as f:
        params = json.load(f)
        
    # 1. Get mapping between mutation ids (from .ssm file) and PyClone cluster ids
    for _, row in pt_ssm_df.iterrows():
        mut_id = row['id']
        mut_items = row['name'].split(":")
        cluster_id = pyclone_df[(pyclone_df['CHR']==int(mut_items[1]))&(pyclone_df['POS']==int(mut_items[2]))&(pyclone_df['REF']==mut_items[3])]['treeCLUSTER'].unique()
        assert(len(cluster_id) <= 1)
        if len(cluster_id) == 1:
            cluster_id = cluster_id.item()
            if cluster_id not in cluster_id_to_mut_ids:
                cluster_id_to_mut_ids[cluster_id] = []
            cluster_id_to_mut_ids[cluster_id].append(mut_id)
            
    # 2. Update params.json with clusters
    clusters = [cluster_id_to_mut_ids[x] for x in range(0, len(cluster_id_to_mut_ids))]
    params['clusters'] = clusters
    with open(os.path.join(output_dir, f"{patient_id}_pyclone_clustered.params.json"), 'w', encoding='utf-8') as f:
        json.dump(params, f, ensure_ascii=False)

def write_pooled_tsv_from_pairtree_clusters(tsv_fn, ssm_fn, clustered_params_json_fn, 
                                            aggregation_rules, output_dir, patient_id,
                                            cluster_sep=";"):

    '''
    After clustering with PairTree (see: https://github.com/morrislab/pairtree#clustering-mutations),
    prepares tsvs by pooling mutations belonging to the same cluster

    Args:
        tsv_fn: path to tsv with reqiured columns: [#sample_index, sample_label, 
        character_index, character_label, anatomical_site_index, anatomical_site_label, 
        ref, var]

        ssm_fn: SSM file used for input into pairtree (see: https://github.com/morrislab/pairtree#input-files)

        clustered_params_json_fn: output params.json file from running PairTree clustervars
        (see: https://github.com/morrislab/pairtree#clustering-mutations)

        aggregation_rules: dictionary indicating how to aggregate any extra columns that are not in 
        the required columns (specificied above). e.g. "first" aggregation rules can be used for columns 
        shared by rows with the same anatomical_site_label, since we are aggregating within an
        anatomical_site_label. 

        output_dir: where to save clustered tsv

        patient_id: name of patient used in tsv filename

        cluster_sep: string that separates names of mutations when creating a cluster name.
        e.g. cluster name for mutations ABC:4:3 and DEF:1:2 with cluster_sep=";" will be 
        "ABC:4:3;DEF:1:2"
    
    Outputs:

        Saves pooled tsv at {output_dir}/{patient_id}_clustered_SNVs.tsv

    '''

    df = pd.read_csv(tsv_fn, delimiter="\t", index_col=0)
    
    # 1. Get mapping between mutation names and pairtree mutation_ids
    mut_name_to_mut_id = dict()
    with open(ssm_fn) as f:
        for i, line in enumerate(f):
            if i == 0: continue
            items = line.split("\t")
            if items[1] not in mut_name_to_mut_id:
                mut_name_to_mut_id[items[1]] = items[0]
                
    # 2. Get pairtree cluster assignments
    with open(clustered_params_json_fn) as f:
        cluster_json = json.loads(f.read())
    mut_name_to_cluster_id = dict()
    for mut_name in df['character_label']:
        cluster_id = _get_cluster_id(mut_name_to_mut_id[mut_name], cluster_json['clusters'])
        mut_name_to_cluster_id[mut_name] = cluster_id

    # 3. Get new names for clustered mutations
    mut_id_to_name = {v:k for k,v in mut_name_to_mut_id.items()}
    cluster_id_to_cluster_name = dict()
    for i, cluster in enumerate(cluster_json['clusters']):
        cluster_comps = []
        for mut in cluster:
            cluster_comps.append(mut_id_to_name[mut])
        cluster_id_to_cluster_name[i] = cluster_sep.join(cluster_comps)

    # 4. Pool mutations and write to file
    dutil.write_pooled_tsv_from_clusters(df, mut_name_to_cluster_id, cluster_id_to_cluster_name,
                                         aggregation_rules, output_dir, patient_id,)

# Adapted from pairtree 
def get_adj_matrix_from_parents(parents):
    K = len(parents) + 1
    T = np.eye(K)
    T[parents,np.arange(1, K)] = 1
    I = np.identity(T.shape[0])
    T = np.logical_xor(T,I).astype(int) # remove self-loops
    # remove the normal subclone
    normal_clone_idx = vutil.get_root_index(T)
    # Assert that this is a monoprimary tree (normal subclone has one cancerous child)
    if len(T[normal_clone_idx].nonzero()[0]) > 1:
        raise ValueError("Normal subclone has more than one child node. Rerun Orchard with monoprimary flag (-p).")

    T = np.delete(T, normal_clone_idx, 0)
    T = np.delete(T, normal_clone_idx, 1)
    return T

def get_adj_matrices_from_pairtree_results(pairtee_results_fns):
    if not isinstance(pairtee_results_fns, list):
        return get_adj_matrix_from_pairtree_results(pairtee_results_fns)

    assert(len(pairtee_results_fns) >= 1)

    return [get_adj_matrix_from_pairtree_results(fn) for fn in pairtee_results_fns]

def get_adj_matrix_from_pairtree_results(pairtee_results_fn):
    results = np.load(pairtee_results_fn)
    parent_vectors = results['struct']
    llhs = results['llh']

    data = []
    for parents_vector, llh in zip(parent_vectors, llhs):
        adj_matrix = get_adj_matrix_from_parents(parents_vector)
        data.append(torch.tensor(adj_matrix, dtype = torch.float32))
    return data




###########################################################################
# projection.py
#
# Contains all of the necessary functions to call the projection algorithm
# from the projectppm library. This is a condensed version of the
# cellular prevalence fitting source from Pairtree that has added documentation.
###########################################################################

import sys, os
import numpy as np
from scipy.stats import binom
import numpy.ctypeslib as npct
import ctypes
from omicsdata.tree.parents import parents_to_adj
from omicsdata.tree.adj import adj_to_anc

from metient.util.globals import MIN_VARIANCE

def _adj_to_adjppm(adj):
    """Convert an adjacency matrix to a format projectppm is expecting
    
    Parameters
    ----------
    adj : ndarray
        an adjaceny matrix for the partial tree

    Returns
    -------
    ndarray 
        an adjacency list used by the project algorithm
    ndarray
        an array containing the degree of each vertex in the tree
    """
    assert np.all(np.diag(adj) == 1) # ensure we have a valid format for an adjacency matrix
    adj = np.copy(adj)
    np.fill_diagonal(adj, 0)
    adjlist = [np.flatnonzero(row) for row in adj + adj.T]
    degrees = np.array([len(row) for row in adjlist], dtype=np.short)
    adjppm = np.zeros_like(adj, dtype=np.short)
    for i, row in enumerate(adjlist):
        adjppm[i, :len(row)] = row
    return adjppm, degrees

def fit_F(parents, V, R, omega, W):
    """Uses the projection algorithm to compute the F matrix
    
    Parameters
    ----------
    parents : ndarray
        a numpy array where each index represents a node, and the value at that index represents
        that nodes direct ancestor (parent) in the tree
    V : ndarray
        a 2D array of variant read counts across all samples for each node. Each row i, V_i, are the 
        variant read counts for the node at index i in the parents array, and each column
        s, V_{is}, are the variant read counts for node i in sample s.
    R : ndarray
        a 2D array of reference read counts across all samples for each node. Each row i, V_i, are the 
        reference read counts for the node at index i in the parents array, and each column
        s, V_{is}, are the reference read counts for node i in sample s.
    omega : ndarray
        a 2D array of variant read probabilities across all samples for each node. Each row i, V_i, are the 
        variant read probabilities for the node at index i in the parents array, and each column
        s, V_{is}, are the variant read probabilities for node i in sample s.
    W : ndarray
        a 2D array of weights used to account for variance in sequencing coverage 
    Returns
    -------
    ndarray
        the cellular prevalence matrix F fit to the (partial) tree
    ndarray
        the subpopulation frequency matrix eta fit to the (partial) tree
    float
        the log binomial likelihood of the cellular prevalence matrix given the read count data
    """
    F, eta = _fit_F(parents, V, R, omega, W)
    F_llh, _, _ = calc_llh(F, V, V + R, omega)
    return F, eta, np.sum(F_llh)

def calc_llh(F, V, N, omega_v, epsilon=1e-5):
    """Computes the log-likelihood under a binomial likelihood model for the cellular prevalence
    matrix F given the read count data
    
    Parameters
    ----------
    F : ndarray
        the mutation frequency matrix F fit to the (partial) tree
    V : ndarray
        a 2D array of variant read counts across all samples for each node. Each row i, V_i, are the 
        variant read counts for the node at index i in the parents array, and each column
        s, V_{is}, are the variant read counts for node i in sample s.
    R : ndarray
        a 2D array of reference read counts across all samples for each node. Each row i, V_i, are the 
        reference read counts for the node at index i in the parents array, and each column
        s, V_{is}, are the reference read counts for node i in sample s.
    omega : ndarray
        a 2D array of variant read probabilities across all samples for each node. Each row i, V_i, are the 
        variant read probabilities for the node at index i in the parents array, and each column
        s, V_{is}, are the variant read probabilities for node i in sample s.
    epsilon : float
        the smallest VAF that a node i can have in a sample s

    Returns
    -------
    ndarray
        a matrix where each entry (i,s) is the log binomial likelihood of the cellluar prevalence of node i in sample s
    ndarray
        an array where each entry (s) is the sum of log binomial likelihoods of the cellluar prevalence all nodes sample s
    float
        the negative log-likelihood of the cellular prevalence matrix
    """
    K, S = F.shape
    for arr in V, N, omega_v:
        assert arr.shape == (K-1, S), "F, V, or N do not have that same input shape (mutation x sample)"

    assert np.allclose(1, F[0]), "The clonal population cellular prevalence is not equal to 1."
    P = omega_v * F[1:]
    P = np.maximum(P, epsilon)
    P = np.minimum(P, 1 - epsilon)

    F_llh = binom.logpmf(V, N, P) 
    assert not np.any(np.isnan(F_llh)) and not np.any(np.isinf(F_llh)), \
           "F log-likelihood is incorrect. Please check inputs to calc_llh."

    llh_per_sample = -np.sum(F_llh, axis=0) / K
    nlglh = np.sum(llh_per_sample) / S
    return (F_llh, llh_per_sample, nlglh)

def _fit_F(parents, V, R, omega, W):
    """Fits the cellular prevalence matrix F matrix one sample at a time
    
    Parameters
    ----------
    parents : ndarray
        a numpy array where each index represents a node, and the value at that index represents
        that nodes direct ancestor (parent) in the tree
    V : ndarray
        a 2D array of variant read counts across all samples for each node. Each row i, V_i, are the 
        variant read counts for the node at index i in the parents array, and each column
        s, V_{is}, are the variant read counts for node i in sample s.
    R : ndarray
        a 2D array of reference read counts across all samples for each node. Each row i, V_i, are the 
        reference read counts for the node at index i in the parents array, and each column
        s, V_{is}, are the reference read counts for node i in sample s.
    omega : ndarray
        a 2D array of variant read probabilities across all samples for each node. Each row i, V_i, are the 
        variant read probabilities for the node at index i in the parents array, and each column
        s, V_{is}, are the variant read probabilities for node i in sample s.
    W : ndarray
        a 2D array of weights used to account for variance in sequencing coverage 
        
    Returns
    -------
    ndarray
        the cellular prevalence matrix F fit to the (partial) tree
    ndarray
        the subpopulation frequency matrix eta fit to the (partial) tree
    """
    # compute total reads (T)
    T = R + V
    M, S = T.shape

    # make an ancestry matrix and adjacency matrix from the parents vector
    adj = parents_to_adj(parents)
    anc = adj_to_anc(adj)
    
    # prevent dividing by zero
    F_hat = np.divide(V, omega * T, out=np.zeros_like(V).astype(np.float64), where=(omega * T)!=0)
    F_hat = np.maximum(0, F_hat)
    F_hat = np.minimum(1, F_hat)
    F_hat = np.insert(F_hat, 0, 1, axis=0)

    V_hat = V + 1
    T_hat = V + R + 2
    W = V_hat*(1 - V_hat/T_hat) / (T_hat*omega)**2
    W = np.maximum(MIN_VARIANCE, W)
    W = np.insert(W, 0, MIN_VARIANCE, axis=0)

    # initialize eta 
    eta = np.zeros((M+1, S))

    # fit eta for each sample
    for sidx in range(S):
        eta[:,sidx] = _fit_eta_S(adj, F_hat[:,sidx], W[:,sidx])
    
    # perform checks
    assert not np.any(np.isnan(eta)), "eta contains np.nans"
    assert np.allclose(0, eta[eta < 0])
    eta[eta < 0] = 0
    assert np.allclose(1, np.sum(eta, axis=0))

    # compute F
    F = np.dot(anc, eta)
    return (F, eta)

def _fit_eta_S(adj, F_hat, var_F_hat, max_retries=3):
    """Fits the eta values for a particular sample using the projection algorithm
    
    Parameters
    ----------
    adj : ndarray
        the adjacency matrix of the (partial) tree
    F_hat : ndarray
        an array of data-implied cellular prevalences for the nodes in the (partial) tree in sample s
    var_F_hat : ndarray
        an array of cellular prevalence variances for each node in sample s
    max_retries : int
        the number of retries to perform if the projection algorithm doesn't provide a valid output

    Returns
    -------
    ndarray
        the array of subpopulation frequency eta for sample s
    """
    M = len(F_hat)
    assert F_hat.ndim == var_F_hat.ndim == 1 and var_F_hat.shape == (M,) and M >=1, \
           "Invalid input"

    # There is an issue with projectppm for certain inputs, see Pairtree source for more information
    for _ in range(max_retries):
        eta = _project_ppm(adj, F_hat, var_F_hat)
        if not np.any(np.isnan(eta)):
            return eta
        print("Projectppm failed to return a valid eta, retrying ...")
    raise Exception('eta still contains NaN after %s attempt(s)' % max_retries)

def _project_ppm(adj, F_hat, var_F_hat, root=0):
    """Wrapper around the C implementation of the projection algorithm. Runs the projection algorithm
    on the data for a single sample (which is how the algorithm is designed).
    
    Parameters
    ----------
    adj : ndarray
        the adjacency matrix of the (partial) tree
    F_hat : ndarray
        an array of data-implied cellular prevalences for the nodes in the (partial) tree in sample s
    var_F_hat : ndarray
        an array of cellular prevalence variances for each node in sample s
    root : int
        the index that designates the root of the tree in the adjacency matrix

    Returns
    -------
    ndarray
        the array of subpopulation frequency eta for sample s
    """
    assert F_hat.ndim == var_F_hat.ndim == 1
    eta = np.empty(len(F_hat), dtype=np.double) # this will be filled out by the call to projectppm

    # see pairtree source for naming conventions
    gamma_init = var_F_hat
    F_hat = F_hat / gamma_init

    # convert the adjacency matrix to a format that projectppm is expecting
    adjppm, degrees = _adj_to_adjppm(adj)

    # make all variables passed into projectppm be C-continguous arrays  
    # see: https://numpy.org/doc/stable/reference/generated/numpy.require.html
    eta = np.require(eta, requirements='C')
    F_hat = np.require(F_hat, requirements='C')
    gamma_init = np.require(gamma_init, requirements='C')
    degrees = np.require(degrees, requirements='C')
    adjppm = np.require(adjppm, requirements='C')

    flag = 0
    compute_eta = 1
    S = 1
    l = len(F_hat)
    # run projectppm projection algorithm, see pairtree source for more documentation on call
    _project_ppm.tree_cost_projection(
        flag, # flag
        compute_eta, # tells to compute eta 
        eta,
        l, # length of input data
        S, # number of samples
        F_hat,
        gamma_init,
        root,
        None,
        None,
        degrees,
        adjppm
    )

    return eta

def _init_projectppm_lib():
    """Initializes all the variables necessary to invoke the projection algorithm from 
    the projectppm library."""

    # define argument types that the projection C function is expecting
    real_arr_1d = npct.ndpointer(dtype=np.float64, ndim=1, flags='C')
    short_arr_1d = npct.ndpointer(dtype=ctypes.c_short, ndim=1, flags='C')
    short_arr_2d = npct.ndpointer(dtype=ctypes.c_short, ndim=2, flags='C')
    class Edge(ctypes.Structure):
        _fields_ = [('first', ctypes.c_short), ('second', ctypes.c_short)]
    c_edge_p = ctypes.POINTER(Edge)
    c_short_p = ctypes.POINTER(ctypes.c_short)
    print(os.path.join(os.path.dirname(__file__)))
    lib_path = os.path.join(os.path.dirname(__file__), 'projectppm', 'bin', 'libprojectppm.so')
    assert os.path.exists(lib_path), 'Could not find projectppm library. \
                                      Please ensure the libprojectppm repository has been cloned in the lib directory,\
                                      and that its source has been built. '

    # loads the projectppm library
    lib = ctypes.cdll.LoadLibrary(lib_path)

    # makes it so ctypes will ensure the argument types 
    # are what the projection function is expecting
    func = lib.tree_cost_projection
    func.argtypes = [
        ctypes.c_short,
        ctypes.c_short,
        real_arr_1d,
        ctypes.c_short,
        ctypes.c_short,
        real_arr_1d,
        real_arr_1d,
        ctypes.c_short,
        c_edge_p,
        c_short_p,
        short_arr_1d,
        short_arr_2d,
    ]
    func.restype = ctypes.c_double
    _project_ppm.tree_cost_projection = func

# Call the init function in here such that each time this file is imported everything we need to 
# use the projection algorithm is already prepared
_init_projectppm_lib() 



import torch
import datetime
import os
import copy
import shutil
import pickle
import gzip
import json

from metient import metient as met
from metient.util import vertex_labeling_util as vutil
from metient.util import data_extraction_util as dutil
from metient.util import eval_util as eutil
from metient.util import plotting_util as putil
from metient.lib import polytomy_resolver as prutil
from metient.lib import v_optimizer as voptim
from metient.lib import u_optimizer as uoptim

from metient.util.globals import *

torch.set_printoptions(precision=2)

def prune_histories(solutions):
    """
    Filter solutions to keep only unique migration histories that lie on the Pareto front with respect to parsimony metrics.
    
    Args:
        solutions (list): List of VertexLabelingSolution objects representing different migration histories
        
    Returns:
        list: Subset of solutions that are unique and Pareto-optimal with respect to migration (m), 
              co-migration (c), and seeding (s) parsimony metrics
    """
    # Collect each unique solution's parsimony metrics
    all_pars_metrics = []
    unique_labelings = set()
    final_solutions = []
    for soln in solutions:
        tree = vutil.MigrationHistory(soln.T, soln.V)
        if tree not in unique_labelings:
            final_solutions.append(soln)
            unique_labelings.add(tree)
            all_pars_metrics.append((soln.m, soln.c, soln.s))
    # Keep the pareto front of trees
    pareto_metrics, pruned_solutions = vutil.pareto_front(final_solutions, all_pars_metrics)
    # print("Pareto_metrics:", set(pareto_metrics))

    return pruned_solutions

def rank_solutions(solution_set, print_config):
    """
    Sort solutions by total loss and return the top k solutions.
    
    Args:
        solution_set (list): List of VertexLabelingSolution objects to rank
        print_config (PrintConfig): Configuration object containing k_best_trees parameter
        
    Returns:
        list: Top k solutions sorted by ascending total loss, where k is specified in print_config.
              If fewer solutions exist than k, returns all solutions.
    """
    # 1. Sort the solutions from lowest to highest loss
    final_solutions = sorted(list(solution_set))

    # 2. Return the k best solutions
    k = print_config.k_best_trees if len(final_solutions) >= print_config.k_best_trees else len(final_solutions)
    final_solutions = final_solutions[:k]
    return final_solutions

def get_best_final_solutions(results, G, O, p, weights, print_config, 
                             node_collection, solve_polytomies, 
                             v_solver, num_internal_nodes, keep_pareto_only=True):
    """
    Process optimization results to get the final set of best migration history solutions.
    
    Args:
        results (list): List of optimization results, each containing:
            - best_Vs: Vertex labeling matrices
            - soft_Vs: Soft vertex labeling matrices (probabilities)
            - best_Ts: Tree adjacency matrices
            - poly_res (PolytomyResolver): Object containing polytomy resolver info
            - metrics: Tuple of (m,c,s,g,o,e) parsimony metrics
        G (torch.Tensor): Genetic distance matrix between sites
        O (dict): Organotropism dictionary mapping sites to frequencies
        p (torch.Tensor): One-hot encoding of primary site
        weights (Weights): Object containing weights for different loss components
        print_config (PrintConfig): Configuration for output formatting
        node_collection (NodeCollection): Object tracking tree node information
        solve_polytomies (bool): Whether polytomy resolution was performed
        v_solver (VertexLabelingSolver): Solver object containing optimization state
        num_internal_nodes (int): Number of internal nodes in tree
        keep_pareto_only (bool, optional): If True, only keep Pareto-optimal solutions. Defaults to True.
        
    Returns:
        list: Ranked list of best VertexLabelingSolution objects after:
            1. Collecting unique solutions
            2. Adding back any removed nodes
            3. Attempting to recover primary-only single-source solutions if none exist
            4. Removing unnecessary polytomy resolver nodes
            5. Computing Pareto front (if keep_pareto_only=True)
            6. Ranking by total loss
    """

    has_pss_solution = False
    full_solution_set = []
    all_pars_metrics, all_result_soln_indices = [],[]
    unique_labelings = set()
        
    # Get the indices of unique solutions
    for result_idx, result in enumerate(results):
        best_Vs, _, best_Ts, _, metrics = result

        for soln_idx, (m,c,s,g,o,e) in enumerate(zip(*metrics)):
            tree = vutil.MigrationHistory(best_Ts[soln_idx].clone(), best_Vs[soln_idx].clone())
            # print((int(m), int(c), int(s)))
            # print(tree.tree, "\n", torch.nonzero(tree.labeling))
            if tree not in unique_labelings:
                all_pars_metrics.append((int(m), int(c), int(s)))
                all_result_soln_indices.append((result_idx, soln_idx, (m,c,s,g,o,e)))
            unique_labelings.add(tree)

    fixed_labeling = v_solver.fixed_labeling
    # Create a list of unique VertexLabelingSolutions
    for idx in all_result_soln_indices:
        result_idx, soln_idx, metrics = idx[0], idx[1], idx[2]

        m,c,s,g,o,e = metrics
        loss = vutil.clone_tree_labeling_loss_with_computed_metrics(m, c, s, g, o, e, weights, bs=1)
        
        # this solution is primary-only single-source, or we've seen one before
        has_pss_solution = s == 1 or has_pss_solution 
        
        V = results[result_idx][0][soln_idx].clone().cpu()
        soft_V = results[result_idx][1][soln_idx].clone().cpu()
        T = results[result_idx][2][soln_idx].clone().cpu()
        # Add back any removed nodes into V and T using FixedVertexLabeling info
        if fixed_labeling is not None and v_solver.config['collapse_nodes']:
            V = add_back_removed_nodes(V, v_solver, p)
            T = add_back_removed_nodes_to_tree(T, v_solver)
            if G is not None:
                G = v_solver.full_G
        soln = vutil.VertexLabelingSolution(loss, m,c,s,g,o,e, V, soft_V, T, G, node_collection)
        full_solution_set.append(soln)
  
    if not has_pss_solution:
        fixed_indices = [x for x in fixed_labeling.known_indices] if fixed_labeling!=None else []
        recover_prim_ss_solutions(full_solution_set, unique_labelings, 
                                  weights, O, p, solve_polytomies, 
                                  fixed_indices, num_internal_nodes)
        
    #  Remove any extra resolver nodes that don't actually help
    poly_res = results[0][3]
    prutil.remove_extra_resolver_nodes(full_solution_set, poly_res, weights, O, p)

    # Compute the pareto front
    if keep_pareto_only:
        pruned_histories = prune_histories(full_solution_set)  
        #pruned_histories = full_solution_set
    else:
        pruned_histories = full_solution_set

    return rank_solutions(pruned_histories, print_config)

def add_back_removed_nodes(V, v_solver, p):
    """Reconstruct full vertex labeling matrix by adding back removed nodes.
    
    Args:
        V (torch.Tensor): Current vertex labeling tensor (num_sites, num_remaining_nodes)
        fixed_labeling (FixedVertexLabeling): Object containing mapping info for removed nodes
        
    Returns:
        torch.Tensor: Expanded vertex labeling tensor including all original nodes
    """
    fixed_labeling = v_solver.fixed_labeling
    if fixed_labeling is None or V.shape[1] == v_solver.full_T.shape[0]:
        return V

    # Get the optimized indices
    optimized_indices_old = [x for x in range(1,v_solver.full_T.shape[0]) if x not in fixed_labeling.known_indices]
    optimized_indices_new = [fixed_labeling.old_to_new[x] for x in optimized_indices_old]

    # Create the full X matrix
    full_V = torch.zeros((V.shape[0], v_solver.full_T.shape[0]), device=V.device)
    full_V[:,0] = p.squeeze()
    # Fill in the full X tensor using the mapped indices
    full_V[:, optimized_indices_old] = V[:, optimized_indices_new]
    # Add in the known labelings (parts of optimal subtrees)
    full_V[:, fixed_labeling.known_indices] = fixed_labeling.known_labelings.to(V.device)

    for x in range(full_V.shape[1]):
        col_sum = int(torch.sum(full_V[:,x]))
        if col_sum != 1:
            print(x, col_sum)

    return full_V

def add_back_removed_nodes_to_tree(T, v_solver):
    """Reconstruct full tree adjacency matrix by adding back removed nodes and their connections.
    
    Args:
        T (torch.sparse.Tensor): Current tree adjacency matrix
        v_solver (VertexLabelingSolver): Object containing info on v optimization
        
    Returns:
        torch.sparse.Tensor: Expanded tree adjacency matrix including all original nodes and connections
    """
    if v_solver.fixed_labeling is None or T.shape[0] == v_solver.full_T.shape[0]:
        return T
        
    # Get dimensions
    num_total_nodes =  v_solver.full_T.shape[0]
    
    # Create mapping from new indices to old indices
    new_to_old = {v: k for k, v in v_solver.fixed_labeling.old_to_new.items()}
    
    # Convert mappings to tensors for vectorized operations
    new_indices = torch.tensor(list(new_to_old.keys()), device=T.device)
    old_indices = torch.tensor(list(new_to_old.values()), device=T.device)
    mapping = torch.zeros(num_total_nodes, dtype=torch.long, device=T.device)
    mapping[new_indices] = old_indices
    
    # Process current T edges
    T = T.coalesce()
    indices = T.indices()
    
    # Map indices using vectorized operations
    mapped_indices = torch.stack([
        mapping[indices[0]],
        mapping[indices[1]]
    ])
    
    # Process full_T edges for removed nodes
    full_T = v_solver.full_T.to_sparse().coalesce().to(T.device)
    full_indices = full_T.indices()
    
    # Create mask for edges involving known nodes
    known_indices = torch.tensor(list(v_solver.fixed_labeling.known_indices), device=T.device)
    known_mask = torch.isin(full_indices[0], known_indices) | torch.isin(full_indices[1], known_indices)
    
    # Combine all indices and values
    all_indices = torch.cat([mapped_indices, full_indices[:, known_mask]], dim=1)
    # Set all values to 1
    all_values = torch.ones(all_indices.shape[1], device=T.device)
    
    # Remove duplicates
    unique_edges, unique_idx = torch.unique(all_indices.T, dim=0, return_inverse=True)
    unique_values = torch.ones_like(unique_edges[:, 0], dtype=torch.float, device=T.device)
    
    # Create final sparse tensor
    new_T = torch.sparse_coo_tensor(
        unique_edges.T, 
        unique_values,
        (num_total_nodes, num_total_nodes)
    ).coalesce()
    
    return new_T

def recover_prim_ss_solutions(solution_set, unique_labelings, 
                              weights, O, p, solve_polytomies, 
                              fixed_indices, num_internal_nodes):
    """
    fixed_indices: indices of nodes that are in optimal subtrees 

    In hard (i.e. usually large input) cases where we are unable to find a 
    primary-only seeding solution, see if we can recover one by post-processing
    final solutions and removing any met-to-met migration edges, and add these
    to our final solution set
    """
    
    new_solutions = []
    for i, solution in enumerate(solution_set):

        #seeding_nodes = putil.seeding_cluster_parents(solution.V,solution.T)
        all_node_indices = [x for x in range(num_internal_nodes)]
        node_info = solution.node_collection
        leaf_nodes = [x for x in all_node_indices if node_info.get_node(x).is_witness]
        # Don't touch the optimal subtrees or leaf nodes
        node_to_label_primary = set(all_node_indices) - (set(fixed_indices).union(set(leaf_nodes)))
        new_V = solution.V.clone()
        for s in node_to_label_primary:
            new_V[:,s] = p.T 
        
        unique_labeled_tree = vutil.MigrationHistory(solution.T, new_V)
        if unique_labeled_tree in unique_labelings:
            continue
        print("calling clone_tree_labeling_objective, updating path matrix", solve_polytomies or i==0)
        loss, new_values = vutil.clone_tree_labeling_objective(new_V, solution.soft_V, solution.T, 
                                                               solution.G, O, p, weights, update_path_matrix=solve_polytomies or i==0, 
                                                               compute_full_c=True)
        new_solution = vutil.VertexLabelingSolution(loss, *new_values, new_V, solution.soft_V, solution.T, solution.G, node_info)
        
        new_solutions.append(new_solution)
        unique_labelings.add(unique_labeled_tree)

    solution_set.extend(new_solutions)
    return

def prep_inputs(tree_fns, tsv_fns, run_names, estimate_observed_clones, output_dir):

    if not (len(tree_fns) == len(tsv_fns) == len(run_names)):
        raise ValueError("Inputs Ts, tsv_fns, and run_names must have equal length (length = number of patients in cohort")

    if isinstance(tree_fns[0], str):
        Ts = []
        for tree_fn in tree_fns:
            Ts.append(dutil.get_adjacency_matrix_from_txt_edge_list(tree_fn))
    else:
        Ts = tree_fns
    
    # If we're not estimating the observed clones, the tsv being inputted doesn't need to be pooled
    if estimate_observed_clones:
        pooled_tsv_fns = []
        for tsv_fn, run_name in zip(tsv_fns, run_names):
            pooled_tsv_fns.append(dutil.pool_input_tsv(tsv_fn, output_dir, f"tmp_{run_name}"))
    else:
        for tsv_fn in tsv_fns:
            dutil.validate_prepooled_tsv(tsv_fn)
        pooled_tsv_fns = tsv_fns
    
    return Ts, pooled_tsv_fns

def one_hot_labeling_for_primary(primary_site, ordered_sites):
    """
    Args:
        - primary_site: name of the primary_site (must be a member of ordered_sites)
        - ordered_sites: list of sites (in the order that they appear in all matrix representations)
    Returns:
        one-hot column vector for a node labeled as belonging to the primary site
    """
    primary_idx = ordered_sites.index(primary_site)
    return torch.nn.functional.one_hot(torch.tensor([primary_idx]), num_classes=len(ordered_sites)).T


def evaluate_label_clone_tree(tree_fn, tsv_fn, weights, print_config, output_dir, run_name, 
             O, sample_size, bias_weights, solve_polytomies, num_runs):
    """
    Observed clone proportions are inputted (in tsv_fns), only labeling of clone tree is needed
    """
    return evaluate(tree_fn, tsv_fn, weights, print_config, output_dir, run_name,
                    O, sample_size, bias_weights, solve_polytomies, num_runs, 
                    estimate_observed_clones=False)

def evaluate(tree_fn, tsv_fn, weights, print_config, output_dir, run_name, 
             O, sample_size, bias_weights, solve_polytomies, num_runs, 
             estimate_observed_clones=True):
    """
    Estimate observed clone proportions and labeling of clone tree for a single patient,
    using inputted weights.
    """
    Ts, pooled_tsv_fns = prep_inputs([tree_fn], [tsv_fn], [run_name], estimate_observed_clones, output_dir)
    assert isinstance(weights.mig, (float, int)), "Weights must be either a float or an int in evaluate mode"
    assert isinstance(weights.seed_site, (float, int)), "Weights must be either a float or an int in evaluate mode"

    T, pooled_tsv_fn = Ts[0], pooled_tsv_fns[0]  
    print(pooled_tsv_fn)  
    primary_sites = dutil.get_primary_sites(pooled_tsv_fn)
    if len(primary_sites) > 1:
        print("Multiple primaries given. Running each as primary")

    for primary_site in primary_sites:
        infer_migration_history(T, pooled_tsv_fn, primary_site, weights, print_config, output_dir, f"{run_name}_{primary_site}", 
                              O=O, sample_size=sample_size, bias_weights=bias_weights, 
                              mode="evaluate", solve_polytomies=solve_polytomies, estimate_observed_clones=estimate_observed_clones,
                              num_runs=num_runs)
    if estimate_observed_clones:
        os.remove(pooled_tsv_fn) # cleanup pooled tsv


def patient_calibration_weight(T, num_poss_primaries):
    """
    Calculate the weight to place on this patient's contribution to the 
    cohort-level cross-entropy score. This is based on the tree size (number of edges)
    and the number of possible primaries, since we don't want to bias towards patients
    with many possible primaries
    """
    num_edges = T.shape[0]-1
    return num_edges/num_poss_primaries

def calibrate_label_clone_tree(tree_fns, tsv_fns, print_config, output_dir, run_names,
                               Os, sample_size, bias_weights, solve_polytomies, num_runs):
    """
    Observed clone proportions are inputted (in tsv_fns), only labeling of clone tree is needed
    """
    return calibrate(tree_fns, tsv_fns, print_config, output_dir, run_names,
                    Os, sample_size, bias_weights, solve_polytomies, num_runs, estimate_observed_clones=False)

def calibrate(tree_fns, tsv_fns, print_config, output_dir, run_names,
              Os, sample_size, bias_weights, solve_polytomies, num_runs,
              estimate_observed_clones=True):
    """
    Estimate observed clone proportions and labeling of clone tree for a cohort of patients,
    calibrate parsimony weights to metastasis priors
    """
    
    Ts, pooled_tsv_fns = prep_inputs(tree_fns, tsv_fns, run_names, estimate_observed_clones, output_dir)

    # Only use maximum parsimony metrics when initially searching for high likelihood trees
    weights = met.Weights(mig=DEFAULT_CALIBRATE_MIG_WEIGHTS, comig=DEFAULT_CALIBRATE_COMIG_WEIGHTS, 
                          seed_site=DEFAULT_CALIBRATE_SEED_WEIGHTS, gen_dist=0.0, 
                          organotrop=0.0)
    # Don't spend time making visualizations for calibrated trees
    visualize = print_config.visualize
    input_k = print_config.k_best_trees
    print_config.visualize = False

    calibrate_dir = os.path.join(output_dir, "calibrate")

    print(f"Saving results to {calibrate_dir}")

    if os.path.exists(calibrate_dir):
        shutil.rmtree(calibrate_dir)
        print(f"Overwriting existing directory at {calibrate_dir}")
    
    os.makedirs(calibrate_dir)
    output_files, patient_weights = [],[]

    # 1. Go through each patient and get migration history in calibrate mode
    for i in range(len(Ts)):
        print("\n*** Calibrating for patient:", run_names[i], "***")
        O = Os[i] if Os != None else None

        primary_sites = dutil.get_primary_sites(pooled_tsv_fns[i])
        if len(primary_sites) > 1:
            print("Multiple primaries given. Running each as primary")

        for primary_site in primary_sites:
            infer_migration_history(Ts[i], pooled_tsv_fns[i], primary_site, weights, print_config, calibrate_dir, f"{run_names[i]}_{primary_site}", 
                                    O=O, sample_size=sample_size, bias_weights=bias_weights,
                                    mode="calibrate", solve_polytomies=solve_polytomies, estimate_observed_clones=estimate_observed_clones,
                                    num_runs=num_runs)
            output_files.append(os.path.join(calibrate_dir, f"{run_names[i]}_{primary_site}.pkl.gz"))
            patient_weights.append(patient_calibration_weight(Ts[i],len(primary_sites)))
            
    # 2. Find the best theta for this cohort
    best_theta = eutil.get_max_cross_ent_thetas(output_files, patient_weights)
    rounded_best_theta = [round(v,3) for v in best_theta]
    with open(os.path.join(calibrate_dir, "best_theta.json"), 'w') as json_file:
        json.dump(rounded_best_theta, json_file, indent=2)

    ordered_sites = dutil.extract_ordered_sites(pooled_tsv_fns)
        
    # 3. Recalibrate trees using the best thetas
    print_config.visualize = visualize
    print_config.k_best_trees = input_k
    cal_weights = met.Weights(mig=[best_theta[0]*50], comig=best_theta[1]*50, seed_site=[best_theta[2]*50],
                          gen_dist=1.0, organotrop=1.0)
    
    # 4. Use the saved trees to rescore trees, visualize, and re-save 
    for i in range(len(Ts)):
        O = Os[i] if Os != None else None
        primary_sites = dutil.get_primary_sites(pooled_tsv_fns[i])
        for primary_site in primary_sites:
            run_name = f"{run_names[i]}_{primary_site}"
            with gzip.open(os.path.join(calibrate_dir, f"{run_name}.pkl.gz"), 'rb') as f:
                pckl = pickle.load(f)

            saved_U = torch.tensor(pckl[OUT_OBSERVED_CLONES_KEY])
            primary_sites = dutil.get_primary_sites(pooled_tsv_fns[i])
            p = one_hot_labeling_for_primary(primary_site, ordered_sites[i])
            reranked_solutions = rank_solutions(vutil.create_reweighted_solution_set_from_pckl(pckl, O, p, cal_weights),
                                                print_config)
            
            putil.save_best_trees(reranked_solutions, saved_U, O, cal_weights, ordered_sites[i], print_config, 
                                  primary_site, calibrate_dir, run_name)
    
    if estimate_observed_clones:
        for pooled_tsv_fn in pooled_tsv_fns:
            os.remove(pooled_tsv_fn) # cleanup pooled tsv

    return best_theta

def to_cpu(tensor):
    """
    Send a tensor to CPU.
    
    Parameters:
        tensor (torch.Tensor or None): The tensor to send to CPU.

    Returns:
        torch.Tensor or None: The tensor on CPU or None if the input was None.
    """
    if tensor is not None:
        return tensor.to('cpu')
    return None

def validate_inputs(T, node_collection, ref, var, primary_site, ordered_sites, weights, O, mode):
    """
    Validate the inputs to Metient.
    """
    if not (T.shape[0] == T.shape[1]):
        raise ValueError(f"Number of tree nodes should be consistent (T.shape[0] == T.shape[1])")
    if (T.shape[0] != len(node_collection.idx_to_label())):
        raise ValueError(f"Number of cluster indices needs to equal shape of adjacency matrix.")
    if not torch.is_tensor(T):
        raise ValueError("T is not a PyTorch tensor.")
    if ref != None and var != None:
        if ref.shape != var.shape:
            raise ValueError(f"ref and var must have identical shape, got {ref.shape} and {var.shape}")
        if not (ref.shape[1] == var.shape[1] == T.shape[0]):
            raise ValueError(f"Number of mutations/mutation clusters should be consistent (ref.shape[1] == var.shape[1] == T.shape[0])")
        if not (ref.shape[0] == var.shape[0] == len(ordered_sites)):   
            raise ValueError(f"Length of ordered_sites should be equal to ref and var dim 0")
    if not vutil.is_tree(T):
        raise ValueError("Adjacency matrix T is empty or not a tree.")
    if not primary_site in ordered_sites:
        raise ValueError(f"{primary_site} not in ordered_sites: {ordered_sites}")
    if (weights.organotrop == 0.0 and O != None)  and mode == "evaluate":
        print(f"Warning: O dictionary was given but organotropism parameter of weights is 0.")
    if (weights.organotrop != 0.0 and O == None):
        raise ValueError(f"O dictionary was not given but organotropism parameter of weights is non-zero. Please pass an O matrix.")
    if mode != 'calibrate' and mode != 'evaluate':
        raise ValueError(f"Valid modes are 'evaluate' and 'calibrate'")
    for label in list(node_collection.idx_to_label().values()):
        if ":" in label:
            raise ValueError(f"Unfortunately our visualization code uses pydot, which does not allow colons (:) in node names. Please use a different separator in 'character_label' values.")


def infer_migration_history(T, tsv_fn, primary_site, weights, print_config, output_dir, run_name, estimate_observed_clones=True,
                            O=None, lr=0.05, init_temp=20, final_temp=0.01, sample_size=-1, bias_weights=True,
                            mode="evaluate", solve_polytomies=False, num_runs=1, keep_pareto_only=True):
    """
    Args:
        T: numpy ndarray or torch tensor (shape: num_internal_nodes x num_internal_nodes). Adjacency matrix (directed) of the internal nodes.
        
        tsv_fn: path to tsv with the required columns: 
            ['anatomical_site_index', 'anatomical_site_label', 'cluster_index', 'character_label', 
            'ref', 'var', 'var_read_prob', 'site_category']

        weights: Weight object for how much to penalize each component of the loss
        
        print_config: PrintConfig object with options on how to visualize output
        
        output_dir: path for where to save output trees to

        run_name: e.g. patient name, used for naming output files.

    Optional:
        
        O: a dictionary mapping anatomical site name (as used in tsv_fn) -> frequency of metastasis (these values should be normalized)

        bias_weights: whether to initialize weights higher to favor vertex labeling of primary + the sites that a node's children are detected in 

        mode: can be "evaluate" or "calibrate"

    Returns:
        Corresponding info on the *best* tree:
        (1) edges of the tree (e.g. [('0', '1'), ('1', '2;3')])
        (2) vertex labeling as a dictionary (e.g. {'0': 'P', '1;3': 'M1'}),
        (3) edges for the migration graph (e.g. [('P', 'M1')])
        (4) dictionary w/ loss values for each component of the loss
        (5) how long (in seconds) the algorithm took to run
    """

    if torch.cuda.is_available():
        torch.set_default_tensor_type(torch.cuda.FloatTensor)
        print("Using GPU")

    start_time = datetime.datetime.now()

    # Extract inputs from tsv
    initialize_G = mode == 'calibrate' or weights.gen_dist != 0
    ref, var, omega, ordered_sites, node_collection, idx_to_observed_sites, G, O = dutil.extract_matrices_from_tsv(tsv_fn, estimate_observed_clones, 
                                                                                                                   T, initialize_G, O, primary_site)
    
    print("Tumor samples:", ordered_sites)
    # Validate inputs
    validate_inputs(T, node_collection, ref, var, primary_site, ordered_sites, weights, O, mode)

    if sample_size == -1:
        sample_size = vutil.calculate_sample_size(T.shape[0], len(ordered_sites), solve_polytomies)
    opt_subtree_sample_size = sample_size
    # Total sample size gets split for the individual parsimony models
    sample_size = sample_size // len(ALL_PARSIMONY_MODELS)
    
    # When calibrating, we want to capture the Pareto front trees and not prune yet
    if mode == 'calibrate':
        print_config.k_best_trees = float("inf")

    # Make the root index 0 (if it isn't already) to simplify indexing logic
    # Save the original root index to swap it back later though, since we don't 
    # want to confuse users with different cluster mappings etc.
    original_root_idx = vutil.get_root_index(T)
    T, ref, var, node_collection, G, idx_to_observed_sites, _, _ = vutil.restructure_matrices_root_index_zero(T, ref, var, node_collection, G, idx_to_observed_sites)
    assert(vutil.get_root_index(T) == 0)

    p = one_hot_labeling_for_primary(primary_site, ordered_sites)
    num_sites = len(ordered_sites)
    num_internal_nodes = T.shape[0]
    
    primary_site_label = ordered_sites[torch.nonzero(p)[0][0]]

    identical_clone_gen_dist = 0.0
    if G != None:
        identical_clone_gen_dist = torch.min(G[(G != 0)])/2.0

    # Keep a copy of input clone tree (T from now on has leaf nodes from U)
    input_T = copy.deepcopy(T)

    # TODO: Make this a more sophisticated decision
    use_sparse_T = input_T.shape[0] > 200

    config = {
        "init_temp": init_temp,
        "final_temp": final_temp,
        "v_anneal_rate": 0.01,
        "t_anneal_rate": 0.01,
        "lr": lr,
        "first_max_iter": 100,
        "second_max_iter": 150 if solve_polytomies else 100,
        "first_v_interval": 15,
        "second_v_interval": 20,
        "opt_subtree_sample_size": opt_subtree_sample_size,
        "sample_size": sample_size,
        "bias_weights": bias_weights,
        "solve_polytomies": solve_polytomies,
        # the genetic distance between two identical clones is a close to 0 but non-zero value
        "identical_clone_gen_dist": identical_clone_gen_dist,
        "num_runs":num_runs,
        "promote_diversity": True,
        "diversity_weight": 0.2,
        "use_sparse_T": use_sparse_T,
        # If we're using a dense T, collapsing nodes is not worth the compute time,
        # and solving polytomies makes the indexing extremely complicated for collapsing (maybe implement in the future)
        "collapse_nodes": use_sparse_T and not solve_polytomies
    }

    ############ Step 1, optimize U ############

    u_optimizer = uoptim.ObservedClonesSolver(num_sites, num_internal_nodes, ref, var, omega, idx_to_observed_sites,
                                              input_T, G, node_collection, weights, config, 
                                              estimate_observed_clones, ordered_sites)
    u_result = u_optimizer.run()
    U, input_T, T, G, L, node_collection, num_internal_nodes, idx_to_observed_sites = u_result

    ############ Step 2, optimize V ############

    num_nodes_to_label = num_internal_nodes - 1 # we don't need to learn the root labeling

    v_optimizer = voptim.VertexLabelingSolver(L, T, p, G, O, weights, config, num_sites, num_nodes_to_label,
                                              node_collection, input_T, idx_to_observed_sites)
    results = v_optimizer.run()
    node_collection = v_optimizer.node_collection

    time_elapsed = (datetime.datetime.now() - start_time).total_seconds()
    if print_config.verbose:
        print(f"Time elapsed: {time_elapsed}")
        
    ############ Step 3, visualize and save outputs ############
    with torch.no_grad():
        # Send tensors to CPU after inference to free up GPU memory
        G = to_cpu(v_optimizer.G)
        O = to_cpu(O)
        p = to_cpu(p)
        
        final_solutions = get_best_final_solutions(results, G, O, p, weights, print_config, 
                                                   node_collection, solve_polytomies,
                                                   v_optimizer, num_internal_nodes, keep_pareto_only=keep_pareto_only)
        print("Number of Pareto-optimal solutions:", len(final_solutions))

        # tracemalloc.start()
        edges, vert_to_site_map, mig_graph_edges, loss_info = putil.save_best_trees(final_solutions, U, O, weights,
                                                                                    ordered_sites, print_config,
                                                                                    primary_site_label, output_dir, run_name,
                                                                                    original_root_idx=original_root_idx) 
        # current, peak = tracemalloc.get_traced_memory()
        # print(f"[After getting pkl outputs ] Current memory usage: {current / 10**6:.2f} MB; Peak was {peak / 10**6:.2f} MB")
        # tracemalloc.stop()

    torch.cuda.empty_cache()
 

    return edges, vert_to_site_map, mig_graph_edges, loss_info, time_elapsed



import torch
from collections import OrderedDict
from metient.util import vertex_labeling_util as vutil
import copy

class PolytomyResolver():

    def __init__(self, v_optimizer, nodes_w_polys, resolver_sites):
        '''
        This is post U matrix estimation, so T already has leaf nodes.
        '''
        
        # 1. Pad the adjacency matrix so that there's room for the new resolver nodes
        # nodes_w_polys are the nodes that have polytomies
        # (we place them in this order: given internal nodes, new resolver nodes, leaf nodes from U)
        T, G = v_optimizer.T, v_optimizer.G

        # TODO: Handle sparse T and G
        # Convert T to dense if it's sparse
        if T.is_sparse:
            T = T.to_dense()

        num_new_nodes = 0
        for r in resolver_sites:
            num_new_nodes += len(r)

        num_leaves = v_optimizer.L.shape[1]
        num_internal_nodes = T.shape[0]-num_leaves
        T = torch.nn.functional.pad(T, pad=(0, num_new_nodes, 0, num_new_nodes), mode='constant', value=0)
        # 2. Shift T and G to make room for the new indices (so the order is input internal nodes, new poly nodes, leaves)
        idx1 = num_internal_nodes
        idx2 = num_internal_nodes+num_leaves
        T = torch.cat((T[:,:idx1], T[:,idx2:], T[:,idx1:idx2]), dim=1)
        if G != None:
            G = torch.nn.functional.pad(G, pad=(0, num_new_nodes, 0, num_new_nodes), mode='constant', value=0)
            G = torch.cat((G[:,:idx1], G[:,idx2:], G[:,idx1:idx2]), dim=1)

        # Shift the leaf node indices in node_collection too
        # This has to be done in descending order otherwise the shifts just overwrite each other
        node_collection = v_optimizer.node_collection
        nodes = node_collection.get_nodes()
        leaf_nodes = [node for node in nodes if node.is_witness]
        leaf_nodes_descending_order = sorted(leaf_nodes, key=lambda obj: obj.idx, reverse=True)
        for node in leaf_nodes_descending_order:
            node_collection.update_index(node.idx, node.idx+num_new_nodes)

        # 3. Get each polytomy's children (these are the positions we have to relearn)
        children_of_polys = vutil.get_child_indices(T, nodes_w_polys)

        # 4. Initialize a matrix to learn the polytomy structure
        ret = initialize_polytomy_resolver_adj_matrix(T, children_of_polys, num_internal_nodes, 
                                                      num_new_nodes, v_optimizer, nodes_w_polys, resolver_sites)
        poly_adj_matrix, nodes_w_polys_to_resolver_indices, resolver_indices, resolver_labeling = ret

        # 5. Initialize potential new nodes as children of the polytomy nodes
        for i in nodes_w_polys:
            for j in nodes_w_polys_to_resolver_indices[i]:
                T[i,j] = 1.0
                parent_node = node_collection.get_node(i)
                new_node = vutil.MigrationHistoryNode(j, [f"{i}pol{j}"]+parent_node.label, is_witness=False, is_polytomy_resolver_node=True)
                node_collection.add_node(new_node)
                if G != None:
                    G[i,j] = v_optimizer.config['identical_clone_gen_dist']

        # 6. The genetic distance between a new node and its potential
        # new children which "switch" is the same distance between the new
        # node's parent and the child
        resolver_index_to_parent_idx = {}
        for poly_node in nodes_w_polys_to_resolver_indices:
            new_nodes = nodes_w_polys_to_resolver_indices[poly_node]
            for new_node_idx in new_nodes:
                resolver_index_to_parent_idx[new_node_idx] = poly_node

        if G != None:
            for new_node_idx in resolver_indices:
                parent_idx = resolver_index_to_parent_idx[new_node_idx]
                potential_child_indices = vutil.get_child_indices(T, [parent_idx])
                for child_idx in potential_child_indices:
                    G[new_node_idx, child_idx] = G[parent_idx, child_idx]
        v_optimizer.T = T
        v_optimizer.G = G
        self.latent_var = poly_adj_matrix
        self.nodes_w_polys = nodes_w_polys
        self.children_of_polys = children_of_polys
        self.resolver_indices = resolver_indices
        self.resolver_index_to_parent_idx = resolver_index_to_parent_idx
        self.resolver_labeling = resolver_labeling

    def _update_indices_with_mapping(self, indices, mapping):
        """Helper function to update indices using a mapping dictionary.
        
        Args:
            indices: List of indices to update
            mapping: Dictionary mapping old indices to new indices
        
        Returns:
            List of updated indices, using original value if not in mapping
        """
        if indices is None:
            return None
        return [mapping[x] if x in mapping else x for x in indices]

    def update_indices(self, old_to_new):
        """Update resolver mappings when nodes are added back to the tree.
        
        Args:
            old_to_new (dict): Mapping from original node indices to reduced tree indices
        """
        # Update nodes with polytomies
        self.nodes_w_polys = self._update_indices_with_mapping(self.nodes_w_polys, old_to_new)
        
        # Update children of polytomies
        self.children_of_polys = self._update_indices_with_mapping(self.children_of_polys, old_to_new)

        # Update resolver indices 
        self.resolver_indices = self._update_indices_with_mapping(self.resolver_indices, old_to_new)
        
        # Update resolver index to parent mapping
        if self.resolver_index_to_parent_idx is not None:
            new_mapping = {}
            for resolver_idx, parent_idx in self.resolver_index_to_parent_idx.items():
                new_resolver_idx = old_to_new[resolver_idx] if resolver_idx in old_to_new else resolver_idx
                new_parent_idx = old_to_new[parent_idx] if parent_idx in old_to_new else parent_idx
                new_mapping[new_resolver_idx] = new_parent_idx
            self.resolver_index_to_parent_idx = new_mapping

def initialize_polytomy_resolver_adj_matrix(T, children_of_polys, num_internal_nodes, 
                                            num_new_nodes, v_optimizer, nodes_w_polys, resolver_sites):
    num_nodes_full_tree = T.shape[0]
    bs = v_optimizer.config['sample_size']
    poly_adj_matrix = vutil.repeat_n(torch.ones((num_nodes_full_tree, len(children_of_polys)), dtype=torch.float32), bs)
    resolver_indices = [x for x in range(num_internal_nodes, num_internal_nodes+num_new_nodes)]

    nodes_w_polys_to_resolver_indices = OrderedDict()
    start_new_node_idx = resolver_indices[0]
    for parent_idx, r in zip(nodes_w_polys, resolver_sites):
        num_new_nodes_for_poly = len(r)
        if parent_idx not in nodes_w_polys_to_resolver_indices:
            nodes_w_polys_to_resolver_indices[parent_idx] = []

        for i in range(start_new_node_idx, start_new_node_idx+num_new_nodes_for_poly):
            nodes_w_polys_to_resolver_indices[parent_idx].append(i)
        start_new_node_idx += num_new_nodes_for_poly

    resolver_labeling = torch.zeros(v_optimizer.num_sites, len(resolver_indices))
    t = 0
    for sites in resolver_sites:
        for site in sites:
            resolver_labeling[site, t] = 1
            t += 1

    offset = 0
    for parent_idx in nodes_w_polys:
        child_indices = vutil.get_child_indices(T, [parent_idx])
        # make the children of polytomies start out as children of their og parent
        # with the option to "switch" to being the child of the new poly node
        poly_adj_matrix[:,parent_idx,offset:(offset+len(child_indices))] = 1.0
        # we only want to let these children choose between being the child
        # of their original parent or the child of this new poly node, which
        # we can do by setting all other indices to -inf
        mask = torch.ones(num_nodes_full_tree, dtype=torch.bool)
        new_nodes = nodes_w_polys_to_resolver_indices[parent_idx]
        mask_indices = new_nodes + [parent_idx]
        mask[[mask_indices]] = 0
        poly_adj_matrix[:,mask,offset:(offset+len(child_indices))] = float("-inf")
        offset += len(child_indices)

    poly_adj_matrix.requires_grad = True

    return poly_adj_matrix, nodes_w_polys_to_resolver_indices, resolver_indices, resolver_labeling
    
def should_remove_node(poly_res, V, T, remove_idx, children_of_removal_node):
    '''

    Returns True if migration graph is the same or better after 
    removing node at index remove_idx

    If any of the following are true:
        (1) the polytomy resolver node is the same color as its parent ,
        (2) the polytomy resolver node only has one child that is the same color as it,
    the migration history won't change by removing the polytomy resolver node. 

    If:
        (3) the polytomy resolver node is the only child of its parent, it's not resolving any polytomies
    
    '''
    parent_idx = poly_res.resolver_index_to_parent_idx[remove_idx]
    remove_idx_color = torch.argmax(V[:,remove_idx]).item()
    is_same_color_as_parent = torch.argmax(V[:,parent_idx]).item() == remove_idx_color
    is_same_color_as_child = torch.argmax(V[:,children_of_removal_node[0]]).item() == remove_idx_color
    # print(parent_idx, remove_idx_color, is_same_color_as_parent, is_same_color_as_child)
    # Case 1
    if is_same_color_as_parent:
        return True
    
    # Case 2
    if len(children_of_removal_node)== 1 and (is_same_color_as_child):
        return True
    
    # Case 3
    children_of_parent_node = vutil.get_child_indices(T, parent_idx)
    if len(children_of_parent_node) == 1:
        return True
    
    return False

def remove_nodes_from_adjacency_matrix_sparse(adj_matrix, removal_nodes):
    '''
    Reindex the adjacency matrix after removing nodes.
    '''
    # Extract indices and values from the sparse adjacency matrix
    adj_matrix = adj_matrix.coalesce()
    indices = adj_matrix.indices()  # Shape: [2, nnz]
    values = adj_matrix.values()   # Shape: [nnz]

    # Convert removal_nodes to a set for fast lookup
    removal_nodes_set = set(removal_nodes)

    # Mask for edges where the target node is in removal_nodes (parents)
    parent_mask = torch.isin(indices[1], torch.tensor(removal_nodes, device=adj_matrix.device))
    parent_nodes = indices[0, parent_mask]
    removal_targets = indices[1, parent_mask]  # The nodes being removed (targets)

    # Mask for edges where the source node is in removal_nodes (children)
    child_mask = torch.isin(indices[0], torch.tensor(removal_nodes, device=adj_matrix.device))
    child_nodes = indices[1, child_mask]
    removal_sources = indices[0, child_mask]  # The nodes being removed (sources)

    # Map each removal node to its parents and children
    removal_to_parents = {}
    removal_to_children = {}

    for target, parent in zip(removal_targets.tolist(), parent_nodes.tolist()):
        if target in removal_to_parents:
            removal_to_parents[target].append(parent)
        else:
            removal_to_parents[target] = [parent]

    for source, child in zip(removal_sources.tolist(), child_nodes.tolist()):
        if source in removal_to_children:
            removal_to_children[source].append(child)
        else:
            removal_to_children[source] = [child]

    # Build new connections: parents -> children for each removal node
    new_edges = []
    for removal_node in removal_nodes_set:
        parents = removal_to_parents.get(removal_node, [])
        children = removal_to_children.get(removal_node, [])

        # Add connections between each parent and child
        for parent in parents:
            for child in children:
                if parent != child:  # Avoid self-loops
                    new_edges.append([parent, child])

    # Convert new edges to a tensor
    new_edges = torch.tensor(new_edges, device=adj_matrix.device).T if new_edges else torch.empty((2, 0), dtype=torch.long, device=adj_matrix.device)
    all_edges = torch.cat((indices, new_edges), dim=1) if new_edges.size(1) > 0 else indices

    # Remove duplicates
    all_edges = torch.unique(all_edges, dim=1)

    # Remove edges that involve removal nodes
    valid_mask = ~(torch.isin(all_edges[0], torch.tensor(removal_nodes, device=adj_matrix.device)) |
                   torch.isin(all_edges[1], torch.tensor(removal_nodes, device=adj_matrix.device)))
    all_edges = all_edges[:, valid_mask]

    # Reindex the nodes
    num_nodes = adj_matrix.size()[0]
    new_index_mapping = {}
    new_index = 0
    for old_index in range(num_nodes):
        if old_index not in removal_nodes_set:
            new_index_mapping[old_index] = new_index
            new_index += 1

    # Map old indices to new indices in edges
    reindexed_edges = torch.stack([
        torch.tensor([new_index_mapping[node.item()] for node in all_edges[0]], device=adj_matrix.device),
        torch.tensor([new_index_mapping[node.item()] for node in all_edges[1]], device=adj_matrix.device)
    ])

    # Create the reindexed sparse adjacency matrix
    new_size = (new_index, new_index)  # The size reduces after removal
    new_values = torch.ones(reindexed_edges.size(1), dtype=values.dtype, device=adj_matrix.device)
    new_adj_matrix = torch.sparse_coo_tensor(reindexed_edges, new_values, size=new_size, dtype=adj_matrix.dtype, device=adj_matrix.device)

    return new_adj_matrix


def remove_nodes_from_adjacency_matrix(T, removal_nodes):
    """
    Reindex the adjacency matrix after removing nodes.
    
    Args:
        T (torch.sparse_coo_tensor): The sparse adjacency matrix.
        remove_nodes (list): A list of node indices to be removed.

    Returns:
        torch.sparse_coo_tensor: The reindexed sparse adjacency matrix.
        dict: Mapping of old indices to new indices.
    """
    if T.is_sparse:
        return remove_nodes_from_adjacency_matrix_sparse(T, removal_nodes)

    T = T.clone().detach()
    # Attach children of the node to remove to their original parent
    for remove_idx in removal_nodes:
        parent_idx = torch.where(T[:,remove_idx] > 0)[0][0]
        child_indices = vutil.get_child_indices(T, [remove_idx])
        for child_idx in child_indices:
            T[parent_idx,child_idx] = 1.0

    # Get the device of the input tensor
    device = T.device
    # Create indices on the same device as adj_matrix
    keep_indices = torch.tensor([i for i in range(T.size(0)) if i not in removal_nodes], device=device)
    # Remove rows of T
    T = T[keep_indices]
    # Remove columns of T
    T = T[:, keep_indices]
    return T


def remove_nodes(removal_indices, V, T, G, node_collection):
    '''
    Remove polytomy resolver nodes from V, T, G and node_idx_to_label
    if they didn't actually help
    '''

    T = remove_nodes_from_adjacency_matrix(T, removal_indices)
    # Remove columns from V
    V = V[:, torch.tensor([i for i in range(V.size(1)) if i not in removal_indices], device=V.device)]

    if G != None: 
        G = G[torch.tensor([i for i in range(G.size(0)) if i not in removal_indices], device=G.device)]
        G = G[:, torch.tensor([i for i in range(G.size(1)) if i not in removal_indices], device=G.device)]
        
    # Reindex the idx to label dict
    node_collection.remove_indices_and_reindex(removal_indices)
    return V, T, G, node_collection

def remove_extra_resolver_nodes(solution_set, poly_res, weights, O, p):
    '''
    If there are any resolver nodes that were added to resolve polytomies but they 
    weren't used (i.e. 1. they have no children or 2. they don't change the 
    migration history), remove them
    '''

    if poly_res == None:
        return solution_set

    for i,soln in enumerate(solution_set):
        modified_soln = None
        
        V, T = soln.V, soln.T
        nodes_to_remove = []
        for new_node_idx in poly_res.resolver_indices:
            children_of_new_node = vutil.get_child_indices(T, new_node_idx)
            if len(children_of_new_node) == 0:
                nodes_to_remove.append(new_node_idx)
            elif should_remove_node(poly_res, V, T, new_node_idx, children_of_new_node):
                nodes_to_remove.append(new_node_idx)
        if len(nodes_to_remove) != 0:
            new_node_collection = copy.deepcopy(soln.node_collection)
            new_V, new_T, new_G, new_node_collection = remove_nodes(nodes_to_remove, V, T, soln.G, new_node_collection)
            loss, new_metrics = vutil.clone_tree_labeling_objective(new_V, soln.soft_V, new_T, new_G, O, p, weights, True)
            modified_soln = vutil.VertexLabelingSolution(loss,*new_metrics,new_V,soln.soft_V,new_T,new_G,new_node_collection)
            solution_set[i] = modified_soln
    
    return


import torch
from metient.util import vertex_labeling_util as vutil

def get_sparse_optimal_subtree_indices(V, P):
    """    
    Parameters:
        V (torch.Tensor): A dense binary tensor (batch_size, num_sites, num_nodes).
        P (torch.sparse.Tensor): A sparse path matrix tensor (batch_size, num_nodes, num_nodes).
        
    Returns:
        torch.Tensor: A tensor resulting from the logical operations.
    """
    bs = P.shape[0]

    cand_optimal_subtree_indices = []
    for i in range(bs):
        _V = V[i]
        _X = (_V.T @ _V)
        _P = P[i].to_dense()
        # i,j = 1 if node i and node j are not connected or they're in the same site
        same_color_subtrees = torch.logical_not(torch.logical_and(1 - _X, _P))
        # Get the indices of rows where all elements are True (all nodes have the same label)
        _optimal_subtree_indices = torch.nonzero(torch.all(same_color_subtrees, dim=1))
        # Tells us how many descendants each node has
        row_sums = torch.sum(_P, dim=1)
        indexed_row_sums = torch.tensor([row_sums[idx[0]] for idx in _optimal_subtree_indices]).unsqueeze(1)
        bs_list = [i for _ in range(len(indexed_row_sums))]
        cand_optimal_subtree_indices.extend([[item[0],int(item[1]),int(item[2])] for item in zip(bs_list, _optimal_subtree_indices, indexed_row_sums)])
        
        del _V, _X, _P, same_color_subtrees
    # Sort the optimal_subtrees by the number of children they have,
    # so that when we are solving for polytomies, we get the largest optimal subtrees possible
    cand_optimal_subtree_indices = sorted(cand_optimal_subtree_indices, key=lambda x: x[2], reverse=True)
    return cand_optimal_subtree_indices

def get_optimal_subtree_indices(T, V, P):

    if T.is_sparse:
        return get_sparse_optimal_subtree_indices(V, P)
    
    VT = torch.transpose(V, 2, 1)
    # i,j = 1 if node i and node j have the same label 
    X = VT @ V
    # i,j = 1 if node i and node j are not connected or they're in the same site
    same_color_subtrees = torch.logical_not(torch.logical_and(1 - X, P))
    # Get the indices of rows where all elements are True (all nodes have the same label)
    cand_optimal_subtree_indices = torch.nonzero(torch.all(same_color_subtrees, dim=2))
    # Tells us how many descendants each node has
    row_sums = torch.sum(P, dim=2)
    indexed_row_sums = torch.tensor([row_sums[idx[0]][idx[1]] for idx in cand_optimal_subtree_indices]).unsqueeze(1)
    cand_optimal_subtree_indices = torch.cat((cand_optimal_subtree_indices, indexed_row_sums), dim=1)
    # 2. Sort the optimal_subtrees by the number of children they have,
    # so that when we are solving for polytomies, we get the largest optimal subtrees possible
    cand_optimal_subtree_indices = cand_optimal_subtree_indices[cand_optimal_subtree_indices[:, 2].argsort(descending=True)]

    return cand_optimal_subtree_indices

def get_descendants(P, batch_num, parent_idx):
    """    
    Parameters:
        P: Path matrix (batch_size, num_nodes, num_nodes).
        batch_num: The batch number to retrieve from
        parent_idx: The index of the parent node whose descendants will be returned 
        
    Returns:
        List of descendant indices of parent_idx (i.e. each descendant can be reached from parent_idx)
    """
    if not P.is_sparse:
        return [t.item() for t in torch.nonzero(P[batch_num,parent_idx])]
        
    P_indices = P.coalesce().indices()
    # Get coalesced indices and create combined mask
    mask = (P_indices[0] == batch_num) & (P_indices[1] == parent_idx)
    
    # Return filtered column indices directly as list
    return P_indices[2][mask].tolist()

def find_optimal_subtree_nodes(T, V, v_solver, num_internal_nodes):
    '''
    Args:
        - T: all possible solutions for adjacency matrices
        - V: all possible solutions for vertex labelings
    Returns:
        A list of node indices and their descendants which belong to optimal subtrees (i.e.)
        all nodes in the subtree have the same color/label, and a list of the batch numbers
        that these optimal subtrees were found 
    '''

    solve_polytomies = v_solver.config['solve_polytomies']
    P = vutil.path_matrix(T, remove_self_loops=False, identical_T=(not solve_polytomies))
    cand_optimal_subtree_indices = get_optimal_subtree_indices(T, V, P)
    seen_nodes = set()
    optimal_batch_nums, optimal_subtree_nodes = [],[]

    for cand in cand_optimal_subtree_indices:
        batch_num = int(cand[0])
        optimal_subtree_root = int(cand[1])
        descendants = set(get_descendants(P, batch_num, optimal_subtree_root))
        # Don't bother with nodes we've already seen
        seen_subset = not optimal_subtree_root in seen_nodes and not (descendants - set([optimal_subtree_root])).issubset(seen_nodes)
        if seen_subset:
            # Add the optimal_subtree_root and all its descendants
            # Don't fix witness nodes (num. descendants == 0), since we already know their labeling, 
            # and if they are under an optimal polytomy branch, they would be getting added by an optimal
            # subtree rooted by an ancestor
            if len(descendants) == 0: 
                continue
            # Don't fix a subtree where there are no leaf nodes 
            # (this is rare, and these nodes aren't well estimated)
            leaf_node_in_optimal_subtree = False
            for descendant in descendants:
                if descendant >= num_internal_nodes:
                    leaf_node_in_optimal_subtree = True
            if not leaf_node_in_optimal_subtree:
                continue

            current_node_set = [optimal_subtree_root]
            seen_nodes.add(optimal_subtree_root)
            for descendant in descendants:
                if descendant not in seen_nodes:
                    current_node_set.append(descendant)
                    seen_nodes.add(descendant)
            if len(current_node_set) > 0:
                optimal_batch_nums.append(batch_num)
                optimal_subtree_nodes.append(current_node_set)
    return optimal_subtree_nodes, optimal_batch_nums


def _collect_node_fixes(node_idx, optimal_batch_num, X, V, T, poly_res):
    """Determine if and how a node's labeling should be fixed.
    
    Args:
        node_idx (int): Index of the node to check
        optimal_batch_num (int): Batch number for this optimal solution
        X (torch.Tensor): Vertex labeling tensor to be modified
        V (torch.Tensor): Current vertex labeling solutions
        T (torch.Tensor): Tree adjacency matrix
        poly_res: Polytomy resolver object or None
    
    Returns:
        tuple: (should_fix, node_idx, optimal_site) or None if node should not be fixed
    """
    node_children = vutil.get_child_indices(T[optimal_batch_num,:,:],[node_idx])
    is_unused_poly_resolver = (poly_res is not None and 
                             node_idx in poly_res.resolver_indices and 
                             len(node_children) < 2)
    
    if (node_idx <= X.shape[2] and node_idx != 0) and not is_unused_poly_resolver:
        optimal_site = int(V[optimal_batch_num,:,node_idx].nonzero(as_tuple=False))
        return (node_idx, optimal_site, optimal_batch_num)
    return None

def _collect_poly_fixes(node_idx, optimal_batch_num, optimal_subtree_set, T, poly_res):
    """Determine if and how a polytomy resolver should be fixed.
    
    Args:
        node_idx (int): Index of the node to check
        optimal_batch_num (int): Batch number for this optimal solution
        optimal_subtree_set (list): Set of nodes in the current optimal subtree
        T (torch.Tensor): Tree adjacency matrix
        poly_res: Polytomy resolver object
        
    Returns:
        tuple: (parent_idx, node_idx, poly_idx, optimal_children) or None if no fixes needed
    """
    if poly_res is None or node_idx not in poly_res.children_of_polys:
        return None
        
    poly_idx = poly_res.children_of_polys.index(node_idx)
    parent_idx = vutil.get_parent(T[optimal_batch_num,:,:], node_idx)
    optimal_children = vutil.get_child_indices(T[optimal_batch_num,:,:], [parent_idx])
    
    if (parent_idx not in optimal_subtree_set or 
        parent_idx not in poly_res.resolver_indices or 
        len(optimal_children) < 2):
        return None
        
    return (parent_idx, node_idx, poly_idx, optimal_children)

def _apply_node_fixes(nodes_to_fix, X, v_solver):
    """Apply collected node labeling fixes to X tensor.
    
    Args:
        nodes_to_fix (list): List of (node_idx, optimal_site, batch_num) tuples
        X (torch.Tensor): Vertex labeling tensor to modify
        v_solver: Vertex labeling solver instance
        
    Returns:
        tuple: (known_indices, known_labelings)
    """
    known_indices = []
    known_labelings = []
    
    for node_idx, optimal_site, _ in nodes_to_fix:
        known_indices.append(node_idx)
        known_labelings.append(torch.eye(v_solver.num_sites)[optimal_site])
        idx = node_idx - 1  # X doesn't include root node
        X[:, optimal_site, idx] = 1
        non_optimal_sites = [i for i in range(v_solver.num_sites) if i != optimal_site]
        X[:, non_optimal_sites, idx] = float("-inf")
        
    return known_indices, known_labelings

def _apply_poly_fixes(poly_fixes, poly_res, T):
    """Apply collected polytomy resolver fixes.
    
    Args:
        poly_fixes (list): List of (parent_idx, node_idx, poly_idx, optimal_children) tuples
        poly_res: Polytomy resolver object
        T (torch.Tensor): Tree adjacency matrix
    """
    poly_resolver_to_optimal_children = {}
    
    for parent_idx, _, poly_idx, optimal_children in poly_fixes:
        # Fix the current child's position
        poly_res.latent_var[:, parent_idx, poly_idx] = 1
        non_parents = [i for i in range(T.shape[1]) if i != parent_idx]
        poly_res.latent_var[:, non_parents, poly_idx] = float("-inf")
        
        # Track optimal children for each resolver
        if parent_idx in poly_res.resolver_indices:
            poly_resolver_to_optimal_children[parent_idx] = optimal_children
            
    return poly_resolver_to_optimal_children

def _remove_nodes_from_T(T, nodes_to_remove):
    T = T.coalesce()
    indices = T.indices()
    values = T.values()
    
    # Create batch-aware mask for nodes to remove
    batch_indices = indices[0]  # (num_edges,)
    source_nodes = indices[1]  # (num_edges,)
    target_nodes = indices[2]  # (num_edges,)
    nodes_to_remove_tensor = torch.tensor(list(nodes_to_remove), device=source_nodes.device)
    # Keep edges where neither source nor target is in nodes_to_remove
    keep_mask = ~((source_nodes.unsqueeze(1) == nodes_to_remove_tensor.unsqueeze(0)).any(dim=1) | 
                    (target_nodes.unsqueeze(1) == nodes_to_remove_tensor.unsqueeze(0)).any(dim=1))
    
    new_indices = indices[:, keep_mask]
    new_values = values[keep_mask]
    # Create mapping for remaining nodes
    remaining_nodes = sorted(set(range(T.shape[1])) - nodes_to_remove)
    old_to_new = {old: new for new, old in enumerate(remaining_nodes)}
    
    # Remap indices
    new_indices[1:] = torch.tensor([
        [old_to_new[idx.item()] for idx in new_indices[1]],
        [old_to_new[idx.item()] for idx in new_indices[2]]
    ])
    
    # Create new T with reduced dimensions
    new_size = (T.shape[0], len(remaining_nodes), len(remaining_nodes))

    return torch.sparse_coo_tensor(new_indices, new_values, new_size).coalesce(), old_to_new

def init_optimal_x_polyres(X, poly_res, optimal_subtree_nodes, optimal_batch_nums, T, V, v_solver):
    """Initialize X and polytomy resolver with optimal subtrees fixed.
    
    This function fixes the labeling of nodes in optimal subtrees and their positions
    in polytomy resolutions. An optimal subtree is one where all nodes have the same label.
    
    Args:
        X (torch.Tensor): Vertex labeling tensor to be modified
        poly_res: Polytomy resolver object or None
        optimal_subtree_nodes (list): Lists of nodes in each optimal subtree
        optimal_batch_nums (list): Batch numbers corresponding to each optimal subtree
        T (torch.Tensor): Tree adjacency matrix
        V (torch.Tensor): Current vertex labeling solutions
        v_solver: Vertex labeling solver instance
        
    Returns:
        tuple: (modified X tensor, modified v_solver)
    """
    # Reset polytomy resolver if it exists
    if poly_res is not None:
        poly_res.latent_var = poly_res.latent_var.detach()
        poly_res.latent_var[poly_res.latent_var != float('-inf')] = 1

    # First pass: collect all modifications needed
    nodes_to_fix = []
    poly_fixes = []
    optimal_root_nodes = []
    
    for optimal_subtree_set, optimal_batch_num in zip(optimal_subtree_nodes, optimal_batch_nums):
        #print("optimal_subtree_set", optimal_subtree_set)
        for i, node_idx in enumerate(optimal_subtree_set):
            # Collect node labeling fixes
            node_fix = _collect_node_fixes(node_idx, optimal_batch_num, X, V, T, poly_res)
            if node_fix:
                nodes_to_fix.append(node_fix)
                if i == 0:
                    optimal_root_nodes.append(node_idx)
                
            # Collect polytomy resolver fixes
            poly_fix = _collect_poly_fixes(node_idx, optimal_batch_num, optimal_subtree_set, T, poly_res)
            if poly_fix:
                poly_fixes.append(poly_fix)

    # Second pass: apply modifications
    known_indices, known_labelings = _apply_node_fixes(nodes_to_fix, X, v_solver)
    #print("known_indices", known_indices)
    if poly_res is not None:
        
        poly_resolver_to_optimal_children = _apply_poly_fixes(poly_fixes, poly_res, T)
        
        # Once a polytomy resolver has found its optimal children, fix its other children so that they cannt move
        for parent_idx, optimal_children in poly_resolver_to_optimal_children.items():
            if len(optimal_children) < 2:
                continue
            optimal_children_poly_indices = [poly_res.children_of_polys.index(i) for i in optimal_children]
            other_children = [i for i in range(poly_res.latent_var.shape[2]) if i not in optimal_children_poly_indices]
            poly_res.latent_var[:, parent_idx, other_children] = float("-inf")

        # Handle resolver indices
        half_bs = X.shape[0]//2
        result = [(x, i) for i, x in enumerate(poly_res.resolver_indices) if x not in known_indices]
        if result:
            unknown_resolver_indices, mask = zip(*result)
            resolved_indices = vutil.repeat_n(poly_res.resolver_labeling[:,mask], half_bs)
            resolved_indices[resolved_indices == 0] = float('-inf')
            X[:half_bs,:,[x-1 for x in unknown_resolver_indices]] = resolved_indices
            
        poly_res.latent_var.requires_grad = True

    # Handle fixed labeling and matrix reduction
    fixed_labeling, old_to_new = None, None
    if known_indices:
        if v_solver.config['collapse_nodes']:
            # Identify which nodes to keep vs remove
            nodes_to_remove = set([x for x in known_indices if x not in optimal_root_nodes])
            # Create new T with reduced dimensions
            _T, old_to_new = _remove_nodes_from_T(T, nodes_to_remove)
            # Don't reformat G if we've already done it in a previous instantiation
            if v_solver.G != None and v_solver.G.shape[0] != _T.shape[1]: 
                G = v_solver.G
                G = G[torch.tensor([i for i in range(G.size(0)) if i not in nodes_to_remove], device=G.device)]
                G = G[:, torch.tensor([i for i in range(G.size(1)) if i not in nodes_to_remove], device=G.device)]
                v_solver.G = G
            v_solver.T = _T[0]
        
        # Update X tensor
        unknown_indices = [x for x in range(v_solver.num_nodes_to_label+1) if x not in known_indices and x != 0]
        known_labelings = torch.stack(known_labelings, dim=1)
        X = X[:,:,[x-1 for x in unknown_indices]]
        #print("num known indices", len(known_indices), "num unknown indices", len(unknown_indices))
        
        fixed_labeling = vutil.FixedVertexLabeling(known_indices, unknown_indices, known_labelings, 
                                                   optimal_root_nodes, old_to_new)

    v_solver.fixed_labeling = fixed_labeling
    X.requires_grad = True

    return X, v_solver

def find_optimal_subtrees(T, V, v_solver):
    '''
    After the first round of optimization, there are optimal subtrees (subtrees where
    the labelings of *all* nodes is the same), which we can keep fixed, since there
    are no other more optimal labelings rooted at this branch.

    Two things we can fix: the labeling of the nodes in optimal subtrees,
    and the edges of the subtrees if polytomy resolution is being used. Search all
    samples to find optimal subtrees, since there might not be one solution with all 
    optimal subtrees.
    '''
    
    num_internal_nodes = v_solver.num_nodes_to_label + 1 # root node
    
    # Re-initialize T with the tree with the best subtree structure
    poly_res = v_solver.poly_res
    if poly_res != None:
        poly_res.latent_var.requires_grad = False
    
    # Find samples with optimal subtrees
    optimal_subtree_nodes, optimal_batch_nums = find_optimal_subtree_nodes(T, V, v_solver, num_internal_nodes)

    return optimal_subtree_nodes, optimal_batch_nums



from setuptools import setup, find_packages
import os
import shutil
from setuptools.command.install import install
from setuptools.command.develop import develop

PACKAGE_DATA_DIR = os.path.join('metient', 'lib', 'projectppm')

def copy_projectppm():
    """Copy projectppm files to package"""
    if not os.path.exists(PACKAGE_DATA_DIR):
        os.makedirs(PACKAGE_DATA_DIR)
        # Copy the files you need
        shutil.copy('projectppm/make.sh', os.path.join(PACKAGE_DATA_DIR, 'make.sh'))
        # Copy other necessary files
        # shutil.copy('projectppm/other_file', os.path.join(PACKAGE_DATA_DIR, 'other_file'))

class CustomInstall(install):
    def run(self):
        copy_projectppm()
        install.run(self)

class CustomDevelop(develop):
    def run(self):
        copy_projectppm()
        develop.run(self)

with open("requirements.txt") as f:
    requirements = [line.strip() for line in f if line.strip() and not line.startswith('#')]

setup(
    name='metient',
    version='0.1.2.3.13',
    url="https://github.com/divyakoyy/metient.git",
    packages=['metient', 'metient.util', 'metient.lib'],
    install_requires=requirements,
    package_data={
        'metient.lib': ['projectppm/*'],  # Include all projectppm files
    },
    include_package_data=True,
    cmdclass={
        'install': CustomInstall,
        'develop': CustomDevelop,
    }
)


import torch
from metient.util import vertex_labeling_util as vutil
from metient.util import data_extraction_util as dutil
from torch.distributions.binomial import Binomial
import numpy as np
import pandas as pd
import numpy as np

from metient.lib.projection import fit_F
from metient.util.globals import MIN_VARIANCE

class ObservedClonesSolver:
    def __init__(self, num_sites, num_internal_nodes, ref, var, omega, idx_to_observed_sites,
                 input_T, G, node_collection, weights, config, estimate_observed_clones, ordered_sites):
        self.ref = ref
        self.var = var
        self.omega = omega
        self.input_T = input_T
        self.G = G
        self.weights = weights
        self.config = config
        self.num_sites = num_sites
        self.num_internal_nodes = num_internal_nodes
        self.node_collection = node_collection
        self.estimate_observed_clones = estimate_observed_clones
        self.idx_to_observed_sites = idx_to_observed_sites
        self.ordered_sites = ordered_sites
    
    def run(self):
        if not self.estimate_observed_clones:
            T, G, L, node_collection = vutil.full_adj_matrix_from_internal_node_idx_to_sites_present(self.input_T, self.G, self.idx_to_observed_sites, 
                                                                                                     self.num_sites, self.config['identical_clone_gen_dist'],
                                                                                                     self.node_collection, self.ordered_sites)
            return None, self.input_T, T, G, L, node_collection, self.num_internal_nodes, self.idx_to_observed_sites

        # return fit_u_map(self)
        return self._fit_u_mle()

    def _fit_u_mle(self):
        """
        Fits the observed clone proportions matrix U using the projection algorithm (finds an MLE estimate of U)
        """
        V, R, omega_V = self.var.T, self.ref.T, self.omega.T
        V_hat = V + 1
        T_hat = V + R + 2
        W = V_hat*(1 - V_hat/T_hat) / (T_hat*omega_V)**2

        # Convert inputs to cpu and numpy float64, as expected by the projection algorithm
        V = V.detach().cpu().numpy().astype(np.float64)
        R = R.detach().cpu().numpy().astype(np.float64)
        W = W.detach().cpu().numpy().astype(np.float64)
        omega = omega_V.detach().cpu().numpy().astype(np.float64)

        W = np.maximum(MIN_VARIANCE, W)

        parents = self._convert_adjmatrix_to_parents()

        F, U, F_llh = fit_F(parents, 
                            V, 
                            R,
                            omega,
                            W)

        U = torch.from_numpy(U).T
        return self._build_full_tree_with_witness_nodes(U)

    def _convert_adjmatrix_to_parents(self):
        """
        Converts the input adjacency matrix to a parent array
        ( a numpy array where each index represents a node, and the value at that index represents
        that nodes direct ancestor (parent) in the tree)
        """
        # Convert input_T to CPU if it's a CUDA tensor
        if torch.is_tensor(self.input_T) and self.input_T.is_cuda:
            input_T = self.input_T.detach().cpu().numpy()
        else:
            input_T = self.input_T
            
        # Add germline root to the tree
        new_T = np.zeros((input_T.shape[0] + 1, input_T.shape[0] + 1))
        new_T[1:, 1:] = input_T
        new_T[0, 1] = 1  # New root connects to old root
        adj = np.copy(new_T)
        np.fill_diagonal(adj, 0)
        return np.argmax(adj[:,1:], axis=0)

    def _build_full_tree_with_witness_nodes(self, U):
        """
        Attaches witness nodes to the inputted tree using the observed clone proportions matrix U

        Args:
            U: Observed clone proportions matrix (num_sites x num_internal_nodes)
            u_solver: ObservedClonesSolver object

        Returns:
            U: Observed clone proportions matrix (num_sites x num_internal_nodes)
            input_T: Input tree (num_internal_nodes x num_internal_nodes)
            T: Expanded tree, which includes internal nodes and leaf/witness nodes (num_tree_nodes x num_tree_nodes)
            G: Genetic distance matrix (num_tree_nodes x num_tree_nodes)
            L: Leaf node labels (num_sites x num_leaf_nodes)
            node_collection: Node collection that contains info on all nodes
            num_internal_nodes: Number of internal nodes
            idx_to_observed_sites: Node index to observed sites
        """
        with torch.no_grad():
            full_T, full_G, L, idx_to_observed_sites, node_collection = vutil.full_adj_matrix_using_inferred_observed_clones(U, self.input_T, self.G, self.num_sites, 
                                                                                                                            self.config['identical_clone_gen_dist'],
                                                                                                                            self.node_collection, self.ordered_sites)
            # Remove any leaf nodes that aren't detected at > U_CUTOFF in any sites. These are not well estimated
            U_no_normal = U[:,1:] # first column is normal cells
            removal_indices = []
            for node_idx in range(U_no_normal.shape[1]):
                children = vutil.get_child_indices(self.input_T, [node_idx])
                if node_idx not in idx_to_observed_sites and len(children) == 0:
                    removal_indices.append(node_idx)
            # print("node indices not well estimated", removal_indices)

            U, input_T, T, G, node_collection, idx_to_observed_sites = vutil.remove_leaf_indices_not_observed_sites(removal_indices, U, self.input_T, 
                                                                                                                    full_T, full_G, node_collection, idx_to_observed_sites)
            num_internal_nodes = self.num_internal_nodes - len(removal_indices)
            
        return U, input_T, T, G, L, node_collection, num_internal_nodes, idx_to_observed_sites


import csv
import numpy as np
import os
import torch
from collections import OrderedDict
import pandas as pd
import sys
from metient.util.globals import *
from metient.util import vertex_labeling_util as vutil

# TODO: make more assertions on uniqueness and completeness of input csvs

def get_adjacency_matrix_from_txt_edge_list(txt_file):
    edges = []
    max_idx = -1
    with open(txt_file) as f:
        for line in f:
            s = line.strip().split(" ")
            edge = (int(s[0]), int(s[1]))
            edges.append(edge)
            max_idx = max(edge[0], max_idx)
            max_idx = max(edge[1], max_idx)

    T = torch.zeros((max_idx+1,max_idx+1))
    for edge in edges:
        T[edge[0], edge[1]] = 1

    return T

def get_mut_to_cluster_map_from_pyclone_output(pyclone_cluster_fn, min_mut_thres=0):
    clstr_id_to_muts, mutation_names = load_pyclone_clusters(pyclone_cluster_fn, min_mut_thres=min_mut_thres)
    mut_name_to_clstr_id = {}
    clstr_id_to_name = {}
    for cid in clstr_id_to_muts:
        for mut in clstr_id_to_muts[cid]:
            mut_name_to_clstr_id[mut] = cid
        clstr_id_to_name[cid] = ";".join(clstr_id_to_muts[cid])
    return mut_name_to_clstr_id, clstr_id_to_name, mutation_names

def _validate_combo(df, tsv_fn, col1, col2, message):
    grouped = df.groupby(col1)[col2].unique()
    valid_mapping = grouped.apply(lambda x: len(x) == 1).all()
    if not valid_mapping:
        raise ValueError(f"{message}. Issue in: {tsv_fn}")
    
def validate_unpooled_or_pooled_df(df, tsv_fn, index_cols):
    '''
    Validation needed for either unpooled or pooled tsv
    '''

    # Some users would like to run on multiple tumors or multiple samples from the same primary tumor 
    assert(set(df['site_category'])==set(['primary', 'metastasis']) or set(df['site_category'])==set(['primary']))

    # Check if all index columns are whole numbers, and convert to int if needed
    for index_col in index_cols:
        assert df[index_col].dtype in ['int64', 'float64'], f"{index_col} column is not int or float"
        if df[index_col].dtype == 'float64':
            is_whole_number = df[index_col].apply(lambda x: x.is_integer())
            has_decimal = not is_whole_number.all()
            if has_decimal:
                raise ValueError(f"{index_col} column must contain whole numbers only. Issue in: {tsv_fn}")
            
    df['cluster_index'] = df['cluster_index'].astype(int)
    
    # Check that each anatomical site index maps to only one anatomical site label
    _validate_combo(df, tsv_fn, 'anatomical_site_index', 'anatomical_site_label', "Each anatomical site index must correspond to the same anatomical site label.")
    # Check that each anatomical site label maps to only one anatomical site index
    _validate_combo(df, tsv_fn, 'anatomical_site_label','anatomical_site_index', "Each anatomical site label must correspond to the same anatomical site index.")
    # Check that each anatomical site index maps to only one site category
    _validate_combo(df, tsv_fn, 'anatomical_site_index', 'site_category', "Each unique anatomical site index must correspond to the same site_category.")
    # Check that all index columns go from 0 to max
    assert set(df['cluster_index']) == set(range(len(df['cluster_index'].unique()))), f"cluster_index values do not go from 0 to max. Issue in: {tsv_fn}"
    assert set(df['anatomical_site_index']) == set(range(len(df['anatomical_site_index'].unique()))), f"anatomical_site_index values do not go from 0 to max, Issue in: {tsv_fn}"

def validate_prepooled_tsv(tsv_fn):
    # Validate the input tsv
    df = pd.read_csv(tsv_fn, delimiter="\t", index_col=False)  
    index_cols = ['anatomical_site_index', 'cluster_index']
    validate_unpooled_or_pooled_df(df, tsv_fn, index_cols)

def pool_input_tsv(tsv_fn, output_dir, run_name):
    '''
    Pool reads from the same anatomical site index and cluster index
    '''
    df = pd.read_csv(tsv_fn, delimiter="\t", index_col=False)  
    validate_unpooled_or_pooled_df(df, tsv_fn, ['anatomical_site_index', 'cluster_index', 'character_index'])
    df['cluster_index'] = df['cluster_index'].astype(int)
    # Check that each character index maps to only one character label
    _validate_combo(df, tsv_fn, 'character_index', 'character_label', "Each character index must correspond to the same character_label.")
    # Check that all var_read_probs are between [0.0,1.0]
    valid_var_read_probs = (df['var_read_prob'] >= 0.0) & (df['var_read_prob'] <= 1.0)
    if not valid_var_read_probs.all():
        raise ValueError(f"All values in the var_read_prob column must be between 0.0 and 1.0 inclusive. Issue in: {tsv_fn}")
    
    _, output_fn = write_pooled_tsv_from_clusters(df, {}, output_dir, run_name)
    
    return output_fn

def write_pooled_tsv_from_clusters(df, aggregation_rules, output_dir, patient_id):
    '''
    After clustering with any clustering algorithm, prepares tsvs by pooling mutations belonging 
    to the same cluster and anatomical site index

    Args:
        df: pandas DataFrame with reqiured columns: [character_label, 
        anatomical_site_index, anatomical_site_label, ref, var, var_read_prob, site_category]

        mut_idx_to_cluster_id: dictionary mapping mutation index ('character_index' in input df)
        to a cluster id

        cluster_id_to_cluster_name: dictionary mapping a cluster id to the name in the output tsv

        aggregation_rules: dictionary indicating how to aggregate any extra columns that are not in 
        the required columns (specificied above). e.g. "first" aggregation rules can be used for columns 
        shared by rows with the same anatomical_site_label, since we are aggregating within an
        anatomical_site_label. 

        output_dir: where to save clustered tsv

        patient_id: name of patient used in tsv filename
    
    Outputs:

        Saves pooled tsv at {output_dir}/{patient_id}_clustered_SNVs.tsv

    '''

    required_cols = ["anatomical_site_index", "anatomical_site_label", "cluster_index",
                     "character_index", "character_label", "ref","var", 
                     "var_read_prob", "site_category"]
    
    all_cols = required_cols + list(aggregation_rules.keys())

    if not (set(required_cols).issubset(df.columns)):
        missing_cols = set(required_cols) - set(df.columns)
        raise ValueError(f"Missing required columns: {missing_cols}")

    if not set(all_cols) == set(df.columns):
        missing_columns = set(df.columns) - set(all_cols)
        df.drop(columns=missing_columns, inplace=True)
        print(f"WARNING: dropping extra columns without aggregation rules: {missing_columns}")

    # 1. Fix the value of each variant's total read count to account for the var_read_prob of each variant,
    # and then we can set var_read_prob to 0.5 (see PairTree's supplement, end of section S3.8)
    # First calculate individual variant read probabilities
    df['total_reads_corrected'] = df.apply(lambda row: 2*row['var_read_prob']*(row['ref']+row['var']), axis=1)
    df['var'] = df.apply(lambda row: min(row['var'], row['total_reads_corrected']), axis=1)
    df['ref'] = df.apply(lambda row: row['total_reads_corrected']-row['var'], axis=1)
    df['var_read_prob'] = 0.5
    df = df.dropna(subset=['cluster_index'])

    # Save the number of mutations in each cluster before pooling
    cluster_id_to_num_muts = df.groupby('cluster_index')['character_index'].nunique().to_dict()
    cluster_id_to_mut_names = df.groupby('cluster_index')['character_label'].unique().apply(list).to_dict()

    # 2. Pool reference and variant allele counts from all mutations within a cluster
    pooled_df = df.drop(['character_label','character_index'], axis=1) # we're going to add this back in later

    ref_var_rules = {'ref': 'sum', 'var': 'sum','total_reads_corrected': 'sum', "var_read_prob": 'first', 
                     'site_category':'first', 'anatomical_site_label':lambda x: ';'.join(set(x)),}

    pooled_df = pooled_df.groupby(['cluster_index', 'anatomical_site_index'], as_index=False).agg({**ref_var_rules, **aggregation_rules})
    # 3. Add indices for mutations, samples and anatomical sites as needed for input format
    pooled_df['character_index'] = pooled_df['cluster_index'].tolist()
    pooled_df['anatomical_site_index'] = pooled_df.apply(lambda row: list(pooled_df['anatomical_site_label'].unique()).index(row["anatomical_site_label"]), axis=1)
    
    # 4. Do some post-processing, e.g. adding number of mutations and shortening character label for display
    pooled_df['num_mutations'] = pooled_df.apply(lambda row:cluster_id_to_num_muts[int(row['character_index'])], axis=1)
    pooled_df['character_label'] = pooled_df.apply(lambda row:cluster_id_to_mut_names[row['cluster_index']], axis=1)
    pooled_df['total_reads_corrected'] = pooled_df['total_reads_corrected'].round(0).astype(int)
    pooled_df['var'] = pooled_df['var'].round(0).astype(int)
    pooled_df['ref'] = pooled_df.apply(lambda row: row['total_reads_corrected']-row['var'], axis=1)
    all_cols.append('num_mutations')

    pooled_df = pooled_df[all_cols]

    # Save
    output_fn = os.path.join(output_dir, f"{patient_id}_clustered_SNVs.tsv")
    with open(output_fn, mode='w', newline='', encoding='utf-8') as file:
        writer = csv.writer(file, delimiter='\t')
        writer.writerow(pooled_df.columns)
        for _, row in pooled_df.iterrows():
            writer.writerow(row)

    return pooled_df, output_fn

def calc_var_read_prob(major_cn, minor_cn, purity, diploid=True):
    major_cn = int(major_cn)
    minor_cn = int(minor_cn)
    p = float(purity)
    factor = 2 if diploid else 1
    x = (p*(major_cn+minor_cn)+factor*(1-p))
    var_read_prob = (major_cn)/x
    return var_read_prob

def get_idx_to_cluster_label(cluster_filepath):
    '''
    Args:
        cluster_filepath: path to cluster file for MACHINA simulated data in the format:
        0
        1
        3;15;17;22;24;29;32;34;53;56
        69;78;80;81

        where each semi-colon separated number represents a mutation that belongs
        to cluster i where i is the file line number.

    Returns:
        (1) a dictionary mapping cluster number to cluster name
        for e.g. for the file above, this would return:
            {0: '0', 1: '69;78;80;81'}
    '''
    idx_to_cluster_label = OrderedDict()
    with open(cluster_filepath) as f:
        i = 0
        for line in f:
            label = line.strip()
            idx_to_cluster_label[i] = label
            i += 1
    return idx_to_cluster_label

def get_ref_var_omega_matrices(tsv_filepaths):
    '''
    tsv_filepaths: List of paths to tsvs (one for each patient). Expects columns:

    ['anatomical_site_index', 'anatomical_site_label', 'character_index', 'character_label', 
    'ref', 'var', 'var_read_prob']
    Returns:
    (1) list of R matrices (num_samples x num_clusters) with the # of reference reads for each sample+cluster,
    (2) list of V matrices (num_samples x num_clusters) with the # of variant reads for each sample+cluster,
    (3) list of omega matrices (num_samples x num_clusters) with the variant read probability of the mutation/mutation cluster in the sample,
    (4) list of list of unique anatomical sites from the patient's data,
    (5) list of dictionaries mapping index to character_label (based on input tsv, gives the index for each mutation name,
    (6) list of dictionaries mapping index to number of mutations (if 'num_mutations' is passed, else None)
    where these indices are used in R matrix, V matrix

    For each of the returned lists, index 0 represents the patient at the first tsv_filepath, etc.
    '''
    if not isinstance(tsv_filepaths, list):
        return get_ref_var_omega_matrix(tsv_filepaths)

    assert(len(tsv_filepaths) >= 1)

    ref_matrices, var_matrices, omega_matrices, ordered_sites, node_collections = [], [], [], [], []
    idx_to_num_muts = None

    for tsv_filepath in tsv_filepaths:
        ref, var, omega, sites, node_info, idx_to_num_mut_dict =  get_ref_var_omega_matrix(tsv_filepath)
        ref_matrices.append(ref)
        var_matrices.append(var)
        omega_matrices.append(omega)
        ordered_sites.append(sites)
        node_collections.append(node_info)
        if idx_to_num_mut_dict != None:
            if idx_to_num_muts == None:
                idx_to_num_muts = []
            idx_to_num_muts.append(idx_to_num_mut_dict)


    return ref_matrices, var_matrices, omega_matrices, ordered_sites, node_collections, idx_to_num_muts

def get_primary_sites(tsv_filepath):
    # For tsvs with very large fields
    csv.field_size_limit(sys.maxsize)

    df = pd.read_csv(tsv_filepath, sep="\t")
    primary_sites = list(df[df['site_category']=='primary']['anatomical_site_label'].unique())
    if len(primary_sites) < 1:
        raise ValueError("No primary found in site_category column")

    return primary_sites

import ast

def get_ref_var_omega_matrix(tsv_filepath):
    '''
    tsv_filepath: path to tsv with columns:

    ['anatomical_site_index', 'anatomical_site_label', 'character_index', 'character_label', 
    'ref', 'var', 'var_read_prob', 'site_category']

    Optional additional flags: 'num_mutations'

    returns
    (1) R matrix (num_samples x num_clusters) with the # of reference reads for each sample+cluster,
    (2) V matrix (num_samples x num_clusters) with the # of variant reads for each sample+cluster,
    (3) omega matrix (num_samples x num_clusters) with the variant read probability of the mutation/mutation cluster in the sample,
    (4) unique anatomical sites from the patient's data,
    (5) dictionary mapping index to mut_names (based on input tsv, gives the index for each mutation name,
    where these indices are used in R matrix, V matrix)
    (6) dictionary mapping index to number of mutations (if 'num_mutations' is passed, else None)
    '''

    df = pd.read_csv(tsv_filepath, delimiter="\t", index_col=False)  

    num_sites = df['anatomical_site_index'].max() + 1
    num_clusters = df['cluster_index'].max() + 1
    # Build R and V matrices
    mig_hist_nodes = []
    R = np.zeros((num_sites, num_clusters))
    V = np.zeros((num_sites, num_clusters))
    omega = np.zeros((num_sites, num_clusters))
    unique_sites = [""]*num_sites
    idx_to_num_muts = dict()
    for _, row in df.iterrows():
        x = int(row['anatomical_site_index'])
        y = int(row['character_index'])
        R[x,y] = int(float(row['ref']))
        V[x,y] = int(float(row['var']))
        omega[x,y] = float(row['var_read_prob'])
        # collect additional metadata
        unique_sites[x] = row['anatomical_site_label']
        # pandas won't let you save a list very easily, so it converts to string...this is a workaround
        # to turn it back into a string
        mig_hist_node = vutil.MigrationHistoryNode(y, ast.literal_eval(row['character_label']), False, False)
        mig_hist_nodes.append(mig_hist_node)
        if y not in idx_to_num_muts:
            idx_to_num_muts[y] = float(row['num_mutations'])
    
    mig_hist_collection = vutil.MigrationHistoryNodeCollection(mig_hist_nodes)

    return torch.tensor(R, dtype=torch.float32), torch.tensor(V, dtype=torch.float32), torch.tensor(omega, dtype=torch.float32), list(unique_sites), mig_hist_collection, idx_to_num_muts

def extract_info_from_observed_clone_tsv(tsv_filename):
    '''
    When observed clones are inputted, we don't need to get ref or var information
    '''
    df = pd.read_csv(tsv_filename, delimiter="\t", index_col=False)  

    num_sites = df['anatomical_site_index'].max() + 1
    mig_hist_nodes = []
    idx_to_sites_present = dict()
    unique_sites = [""]*num_sites
    idx_to_num_muts = dict()
    for _, row in df.iterrows():
        x = int(row['anatomical_site_index'])
        y = int(row['cluster_index'])
        # collect additional metadata
        unique_sites[x] = row['anatomical_site_label']
        mig_hist_node = vutil.MigrationHistoryNode(y, [row['cluster_label']], is_witness=False, is_polytomy_resolver_node=False)
        mig_hist_nodes.append(mig_hist_node)
        if y not in idx_to_num_muts:
            idx_to_num_muts[y] = float(row['num_mutations'])
        if row['present'] == 1:
            if y not in idx_to_sites_present:
                idx_to_sites_present[y] = []
            idx_to_sites_present[y].append(row['anatomical_site_index'])
    mig_hist_collection = vutil.MigrationHistoryNodeCollection(mig_hist_nodes)
    return unique_sites, mig_hist_collection, idx_to_num_muts, idx_to_sites_present

def extract_ordered_sites(tsv_filepaths):
    def _extract_ordered_sites_from_single_tsv(tsv_filename):
        df = pd.read_csv(tsv_filename, delimiter="\t", index_col=False)  
        sorted_df = df.sort_values(by='anatomical_site_index')
        # Get unique labels in the order of index
        unique_site_labels_in_order = sorted_df['anatomical_site_label'].unique().tolist()
        return unique_site_labels_in_order
    
    if not isinstance(tsv_filepaths, list):
        return _extract_ordered_sites_from_single_tsv(tsv_filepaths)

    assert(len(tsv_filepaths) >= 1)
    ordered_sites = []
    for fn in tsv_filepaths:
        ordered_sites.append(_extract_ordered_sites_from_single_tsv(fn))
    return ordered_sites

def initialize_organotropism_vector(O, ordered_sites, primary_site):
    '''
    O: a dictionary mapping anatomical site name (as used in tsv_fn) -> frequency of metastasis (these values should be normalized)
    primary_site: str that represents the primary site
    
    Returns:
    A 1 x num_sites tensor where each value represents the frequency of metastasis to that site
    '''

    # Check for sites in ordered_sites that are not in the dictionary
    for site in ordered_sites:
        if site != primary_site and site not in O:
            raise ValueError(f"Error: '{site}' is not found in the inputted organotropism frequency dictionary.")

    frequency_tensor = torch.zeros(len(ordered_sites), dtype = torch.float32)

    # Fill the tensor with frequencies from the dictionary
    for i, site in enumerate(ordered_sites):
        if site == primary_site:
            frequency_tensor[i] = 0.0  # Set frequency of metastasis to primary site to 0
        else:
            frequency_tensor[i] = O.get(site, 0)  # Get frequency or 0 if site not in dict
    return frequency_tensor

def extract_matrices_from_tsv(tsv_fn, estimate_observed_clones, T, initialize_G, O, primary_site):
    '''
    tsv_fn: path to input tsv
    estimate_observed_clones: whether we need to estimate witness nodes or not
    T: adjacency matrix of clone tree
    initialize_G: whether or not to initialize a genetic distance matrix 
    O: a dictionary mapping anatomical site name (as used in tsv_fn) -> frequency of metastasis (these values should be normalized)
    primary_site: str that represents the primary site
    '''
    if estimate_observed_clones:    
        ref, var, omega, ordered_sites, node_info, idx_to_num_mutations = get_ref_var_omega_matrix(tsv_fn)
        if not torch.is_tensor(ref):
            ref = torch.tensor(ref, dtype=torch.float32)
        if not torch.is_tensor(var):
            var = torch.tensor(var, dtype=torch.float32)
        idx_to_observed_sites = None # needs to be estimated later
    else:
        ref, var, omega = None, None, None
        ordered_sites, node_info, idx_to_num_mutations, idx_to_observed_sites = extract_info_from_observed_clone_tsv(tsv_fn)

    G = None
    # If genetic distance info is given, load into genetic distance matrix
    if idx_to_num_mutations != None and initialize_G:
        G = get_genetic_distance_matrix_from_adj_matrix(T, idx_to_num_mutations)

    organo_vector = None
    if O != None:
        print("Using organotropism frequencies", O)
        organo_vector = initialize_organotropism_vector(O, ordered_sites, primary_site)

    return ref, var, omega, ordered_sites, node_info, idx_to_observed_sites, G, organo_vector

def _get_adj_matrix_from_spruce_tree(tree_edges, idx_to_character_label, remove_unseen_nodes=True):
    '''
    Args:
        tree_edges: list of tuples where each tuple is an edge in the tree
        idx_to_character_label: dictionary mapping index to character_label (machina
        uses colors to represent subclones, so this would map n to 'pink', if pink
        is the nth node in the adjacency matrix).
        remove_unseen_nodes: if True, removes nodes that
        appear in the machina tsv file but do not appear in the reported tree

    Returns:
        T: adjacency matrix where Tij = 1 if there is a path from i to j
        idx_to_character_label: a pruned idx_to_character_label where nodes that
        appear in the machina tsv file but do not appear in the reported tree are removed
    '''
    character_label_to_idx = {v:k for k,v in idx_to_character_label.items()}
    num_internal_nodes = len(character_label_to_idx)
    T = np.zeros((num_internal_nodes, num_internal_nodes))
    seen_nodes = set()
    for edge in tree_edges:
        node_i, node_j = edge[0], edge[1]
        seen_nodes.add(node_i)
        seen_nodes.add(node_j)
        # don't include the leaf/extant nodes (determined from U)
        if node_i in character_label_to_idx and node_j in character_label_to_idx:
            T[character_label_to_idx[node_i], character_label_to_idx[node_j]] = 1

    unseen_nodes = list(set(character_label_to_idx.keys()) - seen_nodes)
    pruned_character_label_to_idx = OrderedDict()
    if remove_unseen_nodes:
        unseen_node_indices = [character_label_to_idx[unseen_node] for unseen_node in unseen_nodes]
        T = np.delete(T, unseen_node_indices, 0)
        T = np.delete(T, unseen_node_indices, 1)

        i = 0
        for char_label in character_label_to_idx:
            if char_label not in unseen_nodes:
                pruned_character_label_to_idx[char_label] = i
                i += 1
        if len(unseen_nodes) > 0:
            #print("Removing unseen nodes:", unseen_nodes, pruned_character_label_to_idx)
            pass

    ret_dict = pruned_character_label_to_idx if remove_unseen_nodes else character_label_to_idx
    ret_dict = {v:k for k,v in ret_dict.items()}
    return T, ret_dict

def get_adj_matrices_from_spruce_mutation_trees(mut_trees_filename, idx_to_character_label, is_sim_data=False):
    '''
    When running MACHINA's generatemutationtrees executable (SPRUCE), it provides a txt file with
    all possible mutation trees. See data/machina_simulated_data/mut_trees_m5/ for examples

    Returns a list of tuples, each containing (T, pruned_idx_to_character_label) for each
    tree in mut_trees_filename.
        - T: adjacency matrix where Tij = 1 if there is a path from i to j
        - idx_to_character_label: a dict mapping indices of the adj matrix T to character
        labels 
    '''
    out = []
    with open(mut_trees_filename, 'r') as f:
        tree_data = []
        for i, line in enumerate(f):
            if i < 3: continue
            # This marks the beginning of a tree
            if "#edges, tree" in line:
                adj_matrix, pruned_idx_to_label = _get_adj_matrix_from_spruce_tree(tree_data, idx_to_character_label)
                out.append((adj_matrix, pruned_idx_to_label))
                tree_data = []
            else:
                nodes = line.strip().split()
                # fixes incompatibility with naming b/w cluster file (uses ";" separator)
                # and the ./generatemutationtrees output (uses "_" separator)
                if is_sim_data:
                    tree_data.append((";".join(nodes[0].split("_")), ";".join(nodes[1].split("_"))))
                else:
                    tree_data.append((nodes[0], nodes[1]))

        adj_matrix, pruned_idx_to_label = _get_adj_matrix_from_spruce_tree(tree_data, idx_to_character_label)
        out.append((adj_matrix, pruned_idx_to_label))
    return out

def get_adj_matrices_from_all_conipher_trees(mut_trees_filename):
    '''
    Extracts trees after running CONIPHER (https://github.com/McGranahanLab/CONIPHER-wrapper)

    Returns a list of adjacency matrices, where Tij = 1 if there is a path from i to j
    '''

    def _get_adj_matrix_from_edges(edges):
        nodes = set([node for edge in edges for node in edge])
        T = np.zeros((len(nodes), len(nodes)))
        for edge in edges:
            T[edge[0], edge[1]] = 1
        return torch.tensor(T, dtype = torch.float32)

    out = []
    with open(mut_trees_filename, 'r') as f:
        tree_edges = []
        for i, line in enumerate(f):
            if i < 2: continue
            # This marks the beginning of a tree
            if "# tree" in line:
                adj_matrix = _get_adj_matrix_from_edges(tree_edges)
                out.append(adj_matrix)
                tree_edges = []
            else:
                nodes = line.strip().split()
                tree_edges.append((int(nodes[0]), int(nodes[1])))

        adj_matrix = _get_adj_matrix_from_edges(tree_edges)
        out.append(adj_matrix)
    return out


def get_genetic_distance_matrices_from_adj_matrices(adj_matrices, idx_to_character_labels, split_char, normalize=True):
    '''
    Get the genetic distances between nodes by counting the number of mutations between
    parent and child. idx_to_character_label's keys are expected to be mutation/cluster indices with
    the mutation or mutations in that cluster (e.g. 'ENAM:4:71507837_DLG1:3:196793590'). split_char
    indicates what the mutations in the cluster name are split by (if it's a cluster).

    If a single adj_matrix or idx_to_character_label is inputted, a single genetic distance matrix is returned. 
    Otherwise, a list of genetic distance matrices is returned.
    '''
    if not isinstance(adj_matrices, list):
        return get_genetic_distance_matrix_from_adj_matrix(adj_matrices, idx_to_character_labels, split_char, normalize=normalize)

    assert(len(adj_matrices) >= 1 and (len(adj_matrices) == len(idx_to_character_labels)))

    gen_dist_matrices = []
    for adj_matrix, idx_to_character_label in zip(adj_matrices, idx_to_character_labels):
        gen_dist_matrix = get_genetic_distance_matrix_from_adj_matrix(adj_matrix, idx_to_character_label, split_char, normalize=normalize)
        gen_dist_matrices.append(gen_dist_matrix)

    return gen_dist_matrices

def get_genetic_distance_matrix_from_adj_matrix(adj_matrix, idx_to_num_muts, normalize=True):
    '''
    Get the genetic distances between nodes by using the number of mutations between parent and child. 
    The number of mutations in a node is given in idx_to_num_muts.
    '''
    G = np.zeros(adj_matrix.shape)

    for i, j in vutil.tree_iterator(adj_matrix):
        # This is the number of mutations the child node has accumulated compared to its parent
        num_mutations = idx_to_num_muts[j]
        G[i][j] = num_mutations

    if normalize:
        G = G / np.sum(G)
    return torch.tensor(G, dtype = torch.float32)

def get_organotropism_matrix_from_msk_met(ordered_sites, cancer_type, frequency_csv, site_to_msk_met_map=None):
    '''
    Args:
        ordered_sites: array of the anatomical site names (e.g. ["breast", "lung"])
        with length =  num_anatomical_sites) where the order matches the order of sites
        in the ref_matrix and var_matrix
        cancer_type: cancer type name, which is used as the 
        frequency_csv: csv with frequency of metastasis by cancer type
        site_to_msk_met_map: dictionary mapping site names to MSK-MET site names. if
        not provided, the names used in the ordered_sites array are used

    Returns:
        array of size len(ordered_sites) with the frequency with which
        primary cancer type seeds site i
    '''

    freq_df = pd.read_csv(frequency_csv)
    if (cancer_type not in list(freq_df['Cancer Type'])):
        raise ValueError(f"{cancer_type} is not in MSK-MET data as a primary cancer type")

    if site_to_msk_met_map != None:
        for site in ordered_sites:
            if (site not in site_to_msk_met_map):
                raise ValueError(f"{site} not in provided site_to_msk_met_map")
        mapped_sites = [site_to_msk_met_map[site] for site in ordered_sites]
    else:
        mapped_sites = ordered_sites

    for site in mapped_sites:
        if (site not in list(freq_df.columns)):
            raise ValueError(f"{site} not in MSK-MET metastatic sites")

    organotrop_arr = np.zeros(len(ordered_sites))
    met_freqs = freq_df[freq_df['Cancer Type']==cancer_type]

    for i, metastatic_site in enumerate(mapped_sites):
        organotrop_arr[i] = met_freqs[metastatic_site].item()

    return torch.tensor(organotrop_arr, dtype = torch.float32)

def load_pyclone_clusters(loci_fn, min_mut_thres=0):
    '''
    Returns (1) map of PyClone cluster ID to mutation names, 
    and (2) list of all mutation names used in PyClone clustering

    Filters clusters with less than min_mut_thres number of mutations
    '''

    # Load map from mutation name to PyClone cluster
    cluster_id_to_mut_names = {}
    cluster_df = pd.read_csv(loci_fn, sep="\t")
    for i, row in cluster_df.iterrows():
        cluster_id = row['cluster_id']
        mutation_id = row['mutation_id']
        if cluster_id not in cluster_id_to_mut_names:
            cluster_id_to_mut_names[cluster_id] = set()
        else:
            cluster_id_to_mut_names[cluster_id].add(mutation_id)
        
    # Filter clusters with less than min_mut_thres mutations
    final_cluster_id_to_mut_names = {}
    mutations = set()
    ix = 0
    for _, v in cluster_id_to_mut_names.items():
        if len(v) >= min_mut_thres:
            mutations.update(v)
            final_cluster_id_to_mut_names[ix] = list(v)
            ix += 1

    return final_cluster_id_to_mut_names, list(mutations)
